# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

#####################################################################
# Bernoulli reconstruction
# Reconstruct an event probability field with values between 0 and 1
# from the observed events
# 1D (set mode=0), 2D (mode=1), or on the sphere (mode=2)
#####################################################################

import numpy as np

import nifty8 as ift


def main():
    # Set up the position space of the signal
    mode = 2
    if mode == 0:
        # One-dimensional regular grid
        position_space = ift.RGSpace(1024)
    elif mode == 1:
        # Two-dimensional regular grid
        position_space = ift.RGSpace([512, 512])
    else:
        # Sphere
        position_space = ift.HPSpace(128)

    # Define harmonic space and transform
    harmonic_space = position_space.get_default_codomain()
    HT = ift.HarmonicTransformOperator(harmonic_space, position_space)

    position = ift.from_random(harmonic_space, 'normal')

    # Define power spectrum and amplitudes
    def sqrtpspec(k):
        return 1./(20. + k**2)

    A = ift.create_power_operator(harmonic_space, sqrtpspec)

    # Set up a sky operator and instrumental response
    sky = ift.sigmoid(HT(A))
    GR = ift.GeometryRemover(position_space)
    R = GR

    # Generate mock data
    p = R(sky)
    mock_position = ift.from_random(harmonic_space, 'normal')
    tmp = p(mock_position).val.astype(np.float64)
    data = ift.random.current_rng().binomial(1, tmp)
    data = ift.Field.from_raw(R.target, data)

    # Compute likelihood energy and Hamiltonian
    position = ift.from_random(harmonic_space, 'normal')
    likelihood_energy = ift.BernoulliEnergy(data) @ p
    ic_newton = ift.DeltaEnergyController(
        name='Newton', iteration_limit=100, tol_rel_deltaE=1e-8)
    minimizer = ift.NewtonCG(ic_newton)
    ic_sampling = ift.GradientNormController(iteration_limit=100)

    # Minimize the Hamiltonian
    H = ift.StandardHamiltonian(likelihood_energy, ic_sampling)
    H = ift.EnergyAdapter(position, H, want_metric=True)
    # minimizer = ift.L_BFGS(ic_newton)
    H, convergence = minimizer(H)

    reconstruction = sky(H.position)

    plot = ift.Plot()
    plot.add(reconstruction, title='reconstruction')
    plot.add(GR.adjoint_times(data), title='data')
    plot.add(sky(mock_position), title='truth')
    plot.output(nx=3, xsize=16, ysize=9, title="results", name="bernoulli.png")


if __name__ == '__main__':
    main()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society
# Author: Philipp Arras

from time import time

import nifty8 as ift
import numpy as np
import pytest

from .common import list2fixture, setup_function, teardown_function

pmp = pytest.mark.parametrize


class NonPureOperator(ift.Operator):
    def __init__(self, domain):
        self._domain = self._target = ift.makeDomain(domain)

    def apply(self, x):
        self._check_input(x)
        return x*time()


class NonPureLinearOperator(ift.LinearOperator):
    def __init__(self, domain, cap):
        self._domain = self._target = ift.makeDomain(domain)
        self._capability = cap

    def apply(self, x, mode):
        self._check_input(x, mode)
        return x*time()


@pmp("cap", [ift.LinearOperator.ADJOINT_TIMES,
             ift.LinearOperator.INVERSE_TIMES | ift.LinearOperator.TIMES])
@pmp("ddtype", [np.float64, np.complex128])
@pmp("tdtype", [np.float64, np.complex128])
def test_purity_check_linear(cap, ddtype, tdtype):
    dom = ift.RGSpace(2)
    op = NonPureLinearOperator(dom, cap)
    with pytest.raises(AssertionError):
        ift.extra.check_linear_operator(op, ddtype, tdtype)


@pmp("dtype", [np.float64, np.complex128])
def test_purity_check(dtype):
    dom = ift.RGSpace(2)
    op = NonPureOperator(dom)
    with pytest.raises(AssertionError):
        ift.extra.check_operator(op, dtype)



# Copyright(C) 2023 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import operator
from collections.abc import Iterable
from functools import partial
from typing import Callable, Tuple, TypeVar, Union

from jax import lax
from jax import numpy as jnp
from jax import random
from jax.tree_util import (tree_leaves, tree_map, tree_structure,
                           tree_transpose, tree_unflatten)

T = TypeVar("T")

CORE_ARITHMETIC_ATTRIBUTES = (
    "__neg__", "__pos__", "__abs__", "__add__", "__radd__", "__sub__",
    "__rsub__", "__mul__", "__rmul__", "__truediv__", "__rtruediv__",
    "__floordiv__", "__rfloordiv__", "__pow__", "__rpow__", "__mod__",
    "__rmod__", "__matmul__", "__rmatmul__"
)


def has_arithmetics(obj, additional_attributes=()):
    desired_attrs = CORE_ARITHMETIC_ATTRIBUTES + additional_attributes
    return all(hasattr(obj, attr) for attr in desired_attrs)


def assert_arithmetics(obj, *args, **kwargs):
    if not has_arithmetics(obj, *args, **kwargs):
        ae = (
            f"input of type {type(obj)} does not support"
            " core arithmetic operations"
            "\nmaybe you forgot to wrap your object in a `Vector`"
        )
        raise AssertionError(ae)


def random_like(key: Iterable, primals, rng: Callable = random.normal):
    import numpy as np

    struct = tree_structure(primals)
    # Cast the subkeys to the structure of `primals`
    subkeys = tree_unflatten(struct, random.split(key, struct.num_leaves))

    def draw(key, x):
        shp = x.shape if hasattr(x, "shape") else jnp.shape(x)
        dtp = x.dtype if hasattr(x, "dtype") else np.result_type(x)
        return rng(key=key, shape=shp, dtype=dtp)

    return tree_map(draw, subkeys, primals)


def unite(x, y, op=operator.add):
    """Unites two Vector-like objects.

    If a key is contained in both objects, then the fields at that key
    are combined.
    """
    from .vector import Vector

    if isinstance(x, Vector) or isinstance(y, Vector):
        x = x.tree if isinstance(x, Vector) else x
        y = y.tree if isinstance(y, Vector) else y
        return Vector(unite(x, y, op=op))
    if not hasattr(x, "keys") and not hasattr(y, "keys"):
        return op(x, y)
    if not hasattr(x, "keys") or not hasattr(y, "keys"):
        te = (
            "one of the inputs does not have a `keys` property;"
            f" got {type(x)} and {type(y)}"
        )
        raise TypeError(te)

    out = {}
    for k in x.keys() | y.keys():
        if k in x and k in y:
            out[k] = op(x[k], y[k])
        elif k in x:
            out[k] = x[k]
        else:
            out[k] = y[k]
    return out


def _shape(x):
    return x.shape if hasattr(x, "shape") else jnp.shape(x)


def tree_shape(tree: T) -> T:
    return tree_map(_shape, tree)


def stack(arrays, axis=0):
    return tree_map(lambda *el: jnp.stack(el, axis=axis), *arrays)


def unstack(stack, axis=0):
    element_count = tree_leaves(stack)[0].shape[0]
    split = partial(jnp.split, indices_or_sections=element_count, axis=axis)
    unstacked = tree_transpose(
        tree_structure(stack), tree_structure((0., ) * element_count),
        tree_map(split, stack)
    )
    return tree_map(partial(jnp.squeeze, axis=axis), unstacked)


def _lax_map(fun, in_axes=0, out_axes=0):
    if in_axes not in (0, (0, )) or out_axes not in (0, (0, )):
        raise ValueError("`lax.map` maps only along first axis")
    return partial(lax.map, fun)


def get_map(map) -> Callable:
    from jax import pmap, vmap

    from ..custom_map import smap, lmap

    if isinstance(map, str):
        if map in ('vmap', 'v'):
            m = vmap
        elif map in ('pmap', 'p'):
            m = pmap
        elif map in ('lmap', 'l'):
            m = lmap
        elif map in ('smap', 's'):
            m = smap
        else:
            raise ValueError(f"unknown `map` {map!r}")
    elif callable(map):
        m = map
    else:
        raise TypeError(f"invalid `map` {map!r}; expected string or callable")
    return m


def map_forest(
    f: Callable,
    in_axes: Union[int, Tuple] = 0,
    out_axes: Union[int, Tuple] = 0,
    tree_transpose_output: bool = True,
    map: Union[str, Callable] = "vmap",
    **kwargs
) -> Callable:
    if out_axes != 0:
        raise NotImplementedError("`out_axis` not yet supported")
    in_axes = in_axes if isinstance(in_axes, tuple) else (in_axes, )
    i = None
    for idx, el in enumerate(in_axes):
        if el is not None and i is None:
            i = idx
        elif el is not None and i is not None:
            nie = "mapping over more than one axis is not yet supported"
            raise NotImplementedError(nie)
    if i is None:
        raise ValueError("must map over at least one axis")
    if not isinstance(i, int):
        te = "mapping over a non integer axis is not yet supported"
        raise TypeError(te)

    map = get_map(map)
    map_f = map(f, in_axes=in_axes, out_axes=out_axes, **kwargs)

    def apply(*xs):
        if not isinstance(xs[i], (list, tuple)):
            te = f"expected mapped axes to be a tuple; got {type(xs[i])}"
            raise TypeError(te)
        x_T = stack(xs[i])

        out_T = map_f(*xs[:i], x_T, *xs[i + 1:])
        # Since `out_axes` is forced to be `0`, we don't need to worry about
        # transposing only part of the output
        if not tree_transpose_output:
            return out_T
        return unstack(out_T)

    return apply


def map_forest_mean(method, map="vmap", *args, **kwargs) -> Callable:
    method_map = map_forest(
        method, *args, tree_transpose_output=False, map=map, **kwargs
    )

    def meaned_apply(*xs, **xs_kw):
        return tree_map(partial(jnp.mean, axis=0), method_map(*xs, **xs_kw))

    return meaned_apply


def mean(forest):
    from functools import reduce
    from .vector import Vector

    norm = 1. / len(forest)
    if isinstance(forest[0], Vector):
        m = norm * reduce(Vector.__add__, forest)
        return m
    else:
        m = norm * reduce(Vector.__add__, (Vector(t) for t in forest))
        return m.tree


def mean_and_std(forest, correct_bias=True):
    from .vector import Vector

    if isinstance(forest[0], Vector):
        m = mean(forest)
        mean_of_sq = mean(tuple(t**2 for t in forest))
    else:
        m = Vector(mean(forest))
        mean_of_sq = mean(tuple(Vector(t)**2 for t in forest))

    n = len(forest)
    scl = jnp.sqrt(n / (n - 1)) if correct_bias else 1.
    std = scl * tree_map(jnp.sqrt, mean_of_sq - m**2)
    if isinstance(forest[0], Vector):
        return m, std
    else:
        return m.tree, std.tree



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import pytest
from numpy.testing import assert_allclose

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize


@pmp('mean', [3., -2])
@pmp('std', [1.5, 0.1])
@pmp('seed', [7, 21])
def test_normal_transform(mean, std, seed):
    op = ift.NormalTransform(mean, std, 'dom')
    assert op.target is ift.DomainTuple.make(())

    op = ift.NormalTransform(mean, std, 'dom', 500)

    with ift.random.Context(seed):
        res = op(ift.from_random(op.domain))
        assert_allclose(res.val.mean(), mean, rtol=0.1)
        assert_allclose(res.val.std(), std, rtol=0.1)

        loc = ift.from_random(op.domain)
        ift.extra.check_operator(op, loc)


@pmp('mean', [0.01, 10.])
@pmp('std_fct', [0.01, 0.1])
@pmp('seed', [7, 21])
def test_lognormal_transform(mean, std_fct, seed):
    std = mean * std_fct
    op = ift.LognormalTransform(mean, std, 'dom', 500)

    with ift.random.Context(seed):
        res = op(ift.from_random(op.domain))
        assert_allclose(res.val.mean(), mean, rtol=0.1)
        assert_allclose(res.val.std(), std, rtol=0.1)

        loc = ift.from_random(op.domain)
        ift.extra.check_operator(op, loc)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
from numpy.testing import assert_allclose

from ..common import setup_function, teardown_function


def test_operator_sum():
    x1 = ift.RGSpace((9,), distances=2.)
    x2 = ift.RGSpace((2, 12), distances=(0.3,))
    dom1 = ift.makeDomain(x1)
    dom2 = ift.makeDomain((x1, x2))
    f1 = ift.from_random(dom1)
    f2 = ift.from_random(dom2)
    arr1 = f1.val
    arr2 = f2.val
    op1 = ift.ScalingOperator(dom1, 1).sum()
    op2 = ift.ScalingOperator(dom2, 1).sum()
    op3 = ift.ScalingOperator(dom2, 1).sum(spaces=1)
    res1 = f1.sum()
    res2 = op1(f1)
    res3 = arr1.sum()
    assert_allclose(res1.val, res2.val)
    assert_allclose(res1.val, res3)
    res4 = f2.sum()
    res5 = op2(f2)
    res6 = arr2.sum()
    assert_allclose(res4.val, res5.val)
    assert_allclose(res4.val, res6)
    res7 = f2.sum(spaces=1)
    res8 = op3(f2)
    res9 = arr2.sum(axis=(1, 2))
    assert_allclose(res7.val, res8.val)
    assert_allclose(res7.val, res9)
    for op in [op1, op2, op3]:
        ift.extra.check_linear_operator(op, domain_dtype=np.float64,
                                        target_dtype=np.float64)
        ift.extra.check_linear_operator(op, domain_dtype=np.complex128,
                                        target_dtype=np.complex128)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest

from ..common import list2fixture, setup_function, teardown_function

pmp = pytest.mark.parametrize


@pmp("op", [ift.LaplaceOperator, ift.UniformOperator])
@pmp("loc", [0, 1.324, 1400])
@pmp("scale", [1.0, 0.0929, 1312.19])
def test_inverse(op, loc, scale):
    dom = ift.UnstructuredDomain([10])
    op = op(dom, loc=loc, scale=scale)
    pos = ift.from_random(dom)
    pos1 = op.inverse(op(pos))
    ift.extra.assert_allclose(pos, pos1)


@pmp("alpha", [1.1, 2, 3, 4])
@pmp("q", [0.2, 1, 5, 10])
@pmp("mode", [0.1, 1])
@pmp("mean", [1.25, 3])
def test_init_parameter_equality(alpha, q, mode, mean):
    dom = ift.UnstructuredDomain([10])
    op1 = ift.InverseGammaOperator(dom, alpha=alpha, q=q)
    op2 = ift.InverseGammaOperator(dom, mode=op1.mode, mean=op1.mean)
    op3 = ift.InverseGammaOperator(dom, mode=mode, mean=mean)
    op4 = ift.InverseGammaOperator(dom, alpha=op3.alpha, q=op3.q)
    f = ift.from_random(dom)
    pos1 = op1(f)
    pos2 = op2(f)
    pos3 = op3(f)
    pos4 = op4(f)
    ift.extra.assert_allclose(pos1, pos2)
    ift.extra.assert_allclose(pos3, pos4)



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from datetime import datetime
from functools import partial
from typing import (
    Any, Callable, Dict, Mapping, NamedTuple, Optional, Tuple, Union
)

from jax import lax
from jax import numpy as jnp
from jax.tree_util import Partial

from . import conjugate_gradient
from .logger import logger
from .tree_math import assert_arithmetics, result_type
from .tree_math import norm as jft_norm
from .tree_math import size, where, vdot


class OptimizeResults(NamedTuple):
    """Object holding optimization results inspired by JAX and scipy.

    Attributes
    ----------
    x : ndarray
        The solution of the optimization.
    success : bool
        Whether or not the optimizer exited successfully.
    status : int
        Termination status of the optimizer. Its value depends on the
        underlying solver. NOTE, in contrast to scipy there is no `message` for
        details since strings are not statically memory bound.
    fun, jac, hess: ndarray
        Values of objective function, its Jacobian and its Hessian (if
        available). The Hessians may be approximations, see the documentation
        of the function in question.
    hess_inv : object
        Inverse of the objective function's Hessian; may be an approximation.
        Not available for all solvers.
    nfev, njev, nhev : int
        Number of evaluations of the objective functions and of its
        Jacobian and Hessian.
    nit : int
        Number of iterations performed by the optimizer.
    """
    x: Any
    success: Union[bool, jnp.ndarray]
    status: Union[int, jnp.ndarray]
    fun: Any
    jac: Any
    hess: Optional[jnp.ndarray] = None
    hess_inv: Optional[jnp.ndarray] = None
    nfev: Union[None, int, jnp.ndarray] = None
    njev: Union[None, int, jnp.ndarray] = None
    nhev: Union[None, int, jnp.ndarray] = None
    nit: Union[None, int, jnp.ndarray] = None
    # Trust-Region specific slots
    trust_radius: Union[None, float, jnp.ndarray] = None
    jac_magnitude: Union[None, float, jnp.ndarray] = None
    good_approximation: Union[None, bool, jnp.ndarray] = None


def _prepare_vag_hessp(fun, jac, hessp,
                       fun_and_grad) -> Tuple[Callable, Callable]:
    """Returns a tuple of functions for computing the value-and-gradient and
    the Hessian-Vector-Product.
    """
    from warnings import warn

    if fun_and_grad is None:
        if fun is not None and jac is not None:
            uw = "computing the function together with its gradient would be faster"
            warn(uw, UserWarning)

            def fun_and_grad(x):
                return (fun(x), jac(x))
        elif fun is not None:
            from jax import value_and_grad

            fun_and_grad = value_and_grad(fun)
        else:
            ValueError("no function specified")

    if hessp is None:
        from jax import grad, jvp

        jac = grad(fun) if jac is None else jac

        def hessp(primals, tangents):
            return jvp(jac, (primals, ), (tangents, ))[1]

    return fun_and_grad, hessp


def newton_cg(fun=None, x0=None, *args, **kwargs):
    """Minimize a scalar-valued function using the Newton-CG algorithm."""
    if x0 is not None:
        assert_arithmetics(x0)
    return _newton_cg(fun, x0, *args, **kwargs).x


def _newton_cg(
    fun=None,
    x0=None,
    *,
    miniter=None,
    maxiter=None,
    energy_reduction_factor=0.1,
    old_fval=None,
    absdelta=None,
    norm_ord=None,
    xtol=1e-5,
    jac: Optional[Callable] = None,
    fun_and_grad=None,
    hessp=None,
    name=None,
    cg=conjugate_gradient._cg,
    cg_kwargs=None,
    time_threshold=None,
    custom_gradnorm=None,
):
    norm_ord = 1 if norm_ord is None else norm_ord
    miniter = 0 if miniter is None else miniter
    maxiter = 200 if maxiter is None else maxiter
    xtol = xtol * size(x0)

    pos = x0
    fun_and_grad, hessp = _prepare_vag_hessp(
        fun, jac, hessp, fun_and_grad=fun_and_grad
    )
    cg_kwargs = {} if cg_kwargs is None else cg_kwargs
    cg_name = name + "CG" if name is not None else None

    gradnorm = (
        partial(jft_norm, ord=norm_ord)
        if custom_gradnorm is None else custom_gradnorm
    )
    energy, g = fun_and_grad(pos)
    nfev, njev, nhev = 1, 1, 0
    if jnp.isnan(energy):
        raise ValueError("energy is Nan")
    status = -1
    i = 0
    for i in range(1, maxiter + 1):
        # Newton approximates the potential up to second order. The CG energy
        # (`0.5 * x.T @ A @ x - x.T @ b`) and the approximation to the true
        # potential in Newton thus live on comparable energy scales. Hence, the
        # energy in a Newton minimization can be used to set the CG energy
        # convergence criterion.
        if old_fval and energy_reduction_factor:
            cg_absdelta = energy_reduction_factor * (old_fval - energy)
        else:
            cg_absdelta = None if absdelta is None else absdelta / 100.
        mag_g = jft_norm(g, ord=cg_kwargs.get("norm_ord", 1))
        cg_resnorm = jnp.minimum(
            0.5, jnp.sqrt(mag_g)
        ) * mag_g  # taken from SciPy
        default_kwargs = {
            "absdelta": cg_absdelta,
            "resnorm": cg_resnorm,
            "norm_ord": 1,
            "_raise_nonposdef": False,  # handle non-pos-def
            "name": cg_name,
            "time_threshold": time_threshold
        }
        cg_res = cg(Partial(hessp, pos), g, **{**default_kwargs, **cg_kwargs})
        nat_g, info = cg_res.x, cg_res.info
        nhev += cg_res.nfev
        if info is not None and info < 0:
            raise ValueError("conjugate gradient failed")

        naive_ls_it = 0
        dd = nat_g  # negative descent direction
        grad_scaling = 1.
        ls_reset = False
        for naive_ls_it in range(9):
            new_pos = pos - grad_scaling * dd
            new_energy, new_g = fun_and_grad(new_pos)
            nfev, njev = nfev + 1, njev + 1
            if new_energy <= energy:
                break

            grad_scaling /= 2
            if naive_ls_it == 5:
                ls_reset = True
                gam = float(vdot(g, g))
                curv = float(g.dot(hessp(pos, g)))
                nhev += 1
                grad_scaling = 1.
                dd = gam / curv * g
        else:
            grad_scaling = 0.
            nm = "N" if name is None else name
            msg = f"{nm}: WARNING: Energy would increase; aborting"
            logger.warning(msg)
            status = -1
            break

        energy_diff = energy - new_energy
        old_fval = energy
        energy = new_energy
        pos = new_pos
        g = new_g

        descent_norm = grad_scaling * gradnorm(dd)
        if name is not None:
            msg = (
                f"{name}: →:{grad_scaling} ↺:{ls_reset} #∇²:{nhev:02d}"
                f" |↘|:{descent_norm:.6e} ➽:{xtol:.6e}"
                f"\n{name}: Iteration {i} ⛰:{energy:+.6e} Δ⛰:{energy_diff:.6e}"
                + (f" ➽:{absdelta:.6e}" if absdelta is not None else "")
            )
            logger.info(msg)
        if jnp.isnan(new_energy):
            raise ValueError("energy is NaN")
        min_cond = naive_ls_it < 2 and i > miniter
        if absdelta is not None and 0. <= energy_diff < absdelta and min_cond:
            status = 0
            break
        if descent_norm <= xtol and i > miniter:
            status = 0
            break
        if time_threshold is not None and datetime.now() > time_threshold:
            status = i
            break
    else:
        status = i
        nm = "N" if name is None else name
        logger.error(f"{nm}: Iteration Limit Reached")
    return OptimizeResults(
        x=pos,
        success=True,
        status=status,
        fun=energy,
        jac=g,
        nit=i,
        nfev=nfev,
        njev=njev,
        nhev=nhev
    )


class _TrustRegionState(NamedTuple):
    x: Any
    converged: Union[bool, jnp.ndarray]
    status: Union[int, jnp.ndarray]
    fun: Any
    jac: Any
    nfev: Union[int, jnp.ndarray]
    njev: Union[int, jnp.ndarray]
    nhev: Union[int, jnp.ndarray]
    nit: Union[int, jnp.ndarray]
    trust_radius: Union[float, jnp.ndarray]
    jac_magnitude: Union[float, jnp.ndarray]
    old_fval: Union[float, jnp.ndarray]


def _trust_ncg(
    fun=None,
    x0=None,
    *,
    maxiter: Optional[int] = None,
    energy_reduction_factor=0.1,
    old_fval=jnp.nan,
    absdelta=None,
    gtol: float = 1e-4,
    max_trust_radius: Union[float, jnp.ndarray] = 1000.,
    initial_trust_radius: Union[float, jnp.ndarray] = 1.0,
    eta: Union[float, jnp.ndarray] = 0.15,
    subproblem=conjugate_gradient._cg_steihaug_subproblem,
    jac: Optional[Callable] = None,
    hessp: Optional[Callable] = None,
    fun_and_grad: Optional[Callable] = None,
    subproblem_kwargs: Optional[Dict[str, Any]] = None,
    name: Optional[str] = None
) -> OptimizeResults:
    from jax.debug import callback

    maxiter = 200 if maxiter is None else maxiter

    status = jnp.where(maxiter == 0, 1, 0)

    if not (0 <= eta < 0.25):
        raise Exception("invalid acceptance stringency")
    # Exception("gradient tolerance must be positive")
    status = jnp.where(gtol < 0., -1, status)
    # Exception("max trust radius must be positive")
    status = jnp.where(max_trust_radius <= 0, -1, status)
    # ValueError("initial trust radius must be positive")
    status = jnp.where(initial_trust_radius <= 0, -1, status)
    # ValueError("initial trust radius must be less than the max trust radius")
    status = jnp.where(initial_trust_radius >= max_trust_radius, -1, status)

    common_dtp = result_type(x0)
    eps = 6. * jnp.finfo(
        common_dtp
    ).eps  # Inspired by SciPy's NewtonCG minimzer

    fun_and_grad, hessp = _prepare_vag_hessp(
        fun, jac, hessp, fun_and_grad=fun_and_grad
    )
    subproblem_kwargs = {} if subproblem_kwargs is None else subproblem_kwargs
    cg_name = name + "SP" if name is not None else None

    f_0, g_0 = fun_and_grad(x0)
    g_0_mag = jft_norm(g_0, ord=subproblem_kwargs.get("norm_ord", 1))
    status = jnp.where(jnp.isfinite(g_0_mag), status, 2)
    init_params = _TrustRegionState(
        converged=False,
        status=status,
        nit=0,
        x=x0,
        fun=f_0,
        jac=g_0,
        jac_magnitude=g_0_mag,
        nfev=1,
        njev=1,
        nhev=0,
        trust_radius=initial_trust_radius,
        old_fval=old_fval
    )

    def pp(arg):
        i = arg["i"]
        msg = (
            "{name}: ↗:{tr:.6e} ⬤:{hit} ∝:{rho:.2e} #∇²:{nhev:02d}"
            "\n{name}: Iteration {i} ⛰:{energy:+.6e} Δ⛰:{energy_diff:.6e}" +
            (" ➽:{absdelta:.6e}" if absdelta is not None else "") +
            ("\n{name}: Iteration Limit Reached" if i == maxiter else "")
        )
        logger.info(msg.format(name=name, **arg))

    def _trust_region_body_f(params: _TrustRegionState) -> _TrustRegionState:
        x_k, g_k, g_k_mag = params.x, params.jac, params.jac_magnitude
        i, f_k, old_fval = params.nit, params.fun, params.old_fval
        tr = params.trust_radius

        i += 1

        if energy_reduction_factor:
            cg_absdelta = energy_reduction_factor * (old_fval - f_k)
        else:
            cg_absdelta = None if absdelta is None else absdelta / 100.
        cg_resnorm = jnp.minimum(0.5, jnp.sqrt(g_k_mag)) * g_k_mag
        # TODO: add an internal success check for future subproblem approaches
        # that might not be solvable
        default_kwargs = {
            "absdelta": cg_absdelta,
            "resnorm": cg_resnorm,
            "trust_radius": tr,
            "norm_ord": 1,
            "name": cg_name
        }
        sub_result = subproblem(
            f_k, g_k, Partial(hessp, x_k), **{
                **default_kwargs,
                **subproblem_kwargs
            }
        )

        pred_f_kp1 = sub_result.pred_f
        x_kp1 = x_k + sub_result.step
        f_kp1, g_kp1 = fun_and_grad(x_kp1)

        actual_reduction = f_k - f_kp1
        pred_reduction = f_k - pred_f_kp1

        # update the trust radius according to the actual/predicted ratio
        rho = actual_reduction / pred_reduction
        tr_kp1 = jnp.where(rho < 0.25, tr * 0.25, tr)
        tr_kp1 = jnp.where(
            (rho > 0.75) & sub_result.hits_boundary,
            jnp.minimum(2. * tr, max_trust_radius), tr_kp1
        )

        # compute norm to check for convergence
        g_kp1_mag = jft_norm(g_kp1, ord=subproblem_kwargs.get("norm_ord", 1))

        # if the ratio is high enough then accept the proposed step
        f_kp1, x_kp1, g_kp1, g_kp1_mag = where(
            rho > eta, (f_kp1, x_kp1, g_kp1, g_kp1_mag),
            (f_k, x_k, g_k, g_k_mag)
        )

        # Check whether we arrived at the float precision
        energy_eps = eps * jnp.abs(f_kp1)
        converged = (actual_reduction
                     <= energy_eps) & (actual_reduction > -energy_eps)

        converged |= g_kp1_mag < gtol
        if absdelta:
            converged |= (rho > eta) & (actual_reduction
                                        > 0.) & (actual_reduction < absdelta)

        status = jnp.where(converged, 0, params.status)
        status = jnp.where(i >= maxiter, 1, status)
        status = jnp.where(pred_reduction <= 0, 2, status)
        params = _TrustRegionState(
            converged=converged,
            nit=i,
            x=x_kp1,
            fun=f_kp1,
            jac=g_kp1,
            jac_magnitude=g_kp1_mag,
            nfev=params.nfev + sub_result.nfev + 1,
            njev=params.njev + sub_result.njev + 1,
            nhev=params.nhev + sub_result.nhev,
            trust_radius=tr_kp1,
            status=status,
            old_fval=f_k
        )
        if name is not None:
            printable_state = {
                "i": i,
                "energy": params.fun,
                "energy_diff": actual_reduction,
                "maxiter": maxiter,
                "absdelta": absdelta,
                "tr": params.trust_radius,
                "rho": rho,
                "nhev": params.nhev,
                "hit": sub_result.hits_boundary
            }
            callback(pp, printable_state)
        return params

    def _trust_region_cond_f(params: _TrustRegionState) -> bool:
        return jnp.logical_not(params.converged) & (params.status == 0)

    state = lax.while_loop(
        _trust_region_cond_f, _trust_region_body_f, init_params
    )

    return OptimizeResults(
        success=state.converged & (state.status == 0),
        nit=state.nit,
        x=state.x,
        fun=state.fun,
        jac=state.jac,
        nfev=state.nfev,
        njev=state.njev,
        nhev=state.nhev,
        jac_magnitude=state.jac_magnitude,
        trust_radius=state.trust_radius,
        status=state.status
    )


def trust_ncg(fun=None, x0=None, *args, **kwargs):
    if x0 is not None:
        assert_arithmetics(x0)
    return _trust_ncg(fun, x0, *args, **kwargs).x


def minimize(
    fun: Optional[Callable[..., float]],
    x0,
    args: Tuple = (),
    *,
    method: str,
    tol: Optional[float] = None,
    options: Optional[Mapping[str, Any]] = None
) -> OptimizeResults:
    """Minimize fun."""
    assert_arithmetics(x0)
    if options is None:
        options = {}
    if not isinstance(args, tuple):
        te = f"args argument must be a tuple, got {type(args)!r}"
        raise TypeError(te)

    fun_with_args = fun
    if args:
        fun_with_args = lambda x: fun(x, *args)

    if tol is not None:
        raise ValueError("use solver-specific options")

    if method.lower() in ('newton-cg', 'newtoncg', 'ncg'):
        return _newton_cg(fun_with_args, x0, **options)
    elif method.lower() in ('trust-ncg', 'trustncg'):
        return _trust_ncg(fun_with_args, x0, **options)

    raise ValueError(f"method {method} not recognized")



# Copyright(C) 2013-2021 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from .lanczos import (
    lanczos_tridiag, stochastic_logdet_from_lanczos, stochastic_lq_logdet
)
from .stats_distributions import (
    interpolator, invgamma_invprior, laplace_prior, invgamma_prior,
    lognormal_invprior, lognormal_moments, lognormal_prior, normal_invprior,
    normal_prior, uniform_prior
)
from .unique import amend_unique, amend_unique_, unique



import pytest

pytest.importorskip("jax")

import sys

from jax import grad
from jax import numpy as jnp
from numpy.testing import assert_allclose

import nifty8.re as jft

pmp = pytest.mark.parametrize

pot_and_tol = (
    (
        lambda q: jnp.
        sum(q.T @ jnp.linalg.inv(jnp.array([[1, 0.95], [0.95, 1]])) @ q / 2.),
        0.2
    ), (lambda q: -1 / jnp.linalg.norm(q), 2e-2)
)


@pmp("potential_energy, rtol", pot_and_tol)
def test_leapfrog_energy_conservation(
    potential_energy, rtol, interactive=False
):
    dims = (2, )
    mass_matrix = jnp.ones(shape=dims)
    kinetic_energy = lambda p: jnp.sum(p**2 / mass_matrix / 2.)

    potential_energy_gradient = grad(potential_energy)
    positions = [jnp.array([-1.5, -1.55])]
    momenta = [jnp.array([-1, 1])]
    for _ in range(25):
        new_qp = jft.hmc.leapfrog_step(
            qp=jft.hmc.QP(position=positions[-1], momentum=momenta[-1]),
            potential_energy_gradient=potential_energy_gradient,
            kinetic_energy_gradient=lambda x, y: x * y,
            step_size=0.25,
            inverse_mass_matrix=1. / mass_matrix
        )
        positions.append(new_qp.position)
        momenta.append(new_qp.momentum)

    potential_energies = list(map(potential_energy, positions))
    kinetic_energies = list(map(kinetic_energy, momenta))

    jnp.set_printoptions(precision=2)
    for q, p, e_kin, e_pot in zip(
        positions, momenta, potential_energies, kinetic_energies
    ):
        msg = (
            f"q: {q}; p: {p}"
            f"\nE_tot: {e_pot+e_kin:.2e}; E_pot: {e_pot:.2e}; E_kin: {e_kin:.2e}"
        )
        print(msg, file=sys.stderr)

    old_energy_tot = potential_energies[0] + kinetic_energies[0]
    new_energy_tot = potential_energies[-1] + kinetic_energies[-1]
    assert_allclose(old_energy_tot, new_energy_tot, rtol=rtol)

    if interactive:
        return positions, momenta, kinetic_energies, potential_energies


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    qs, ps, e_kins, e_pots = test_leapfrog_energy_conservation(
        *pot_and_tol[0], interactive=True
    )
    positions = jnp.array(qs)
    momenta = jnp.array(ps)
    kinetic_energies = jnp.array(e_kins)
    potential_energies = jnp.array(e_pots)

    # Position Coordinates
    plt.plot(positions[:, 0], positions[:, 1])
    plt.xlabel("position[:,0]")
    plt.ylabel("position[:,1]")
    plt.show()

    # Momentum coordinates
    plt.plot(momenta[:, 0], momenta[:, 1])
    plt.xlabel("momenta[:,0]")
    plt.ylabel("momenta[:,1]")
    plt.show()

    # Value of Hamiltonian
    # does not look exactly the same as in Neal (2011) unfortunately!
    plt.plot(kinetic_energies, label='kin')
    plt.plot(potential_energies, label='pot')
    plt.plot(kinetic_energies + potential_energies, label='total')
    plt.xlabel('time')
    plt.ylabel('energy')
    plt.legend()
    plt.show()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np
from scipy.sparse import coo_matrix
from scipy.sparse.linalg import aslinearoperator
from scipy.special import erfc

from ..domain_tuple import DomainTuple
from ..domains.rg_space import RGSpace
from ..domains.unstructured_domain import UnstructuredDomain
from ..field import Field
from ..operators.linear_operator import LinearOperator


def _gaussian_sf(x):
    return 0.5*erfc(x/np.sqrt(2.))


def _comp_traverse(start, end, shp, dist, lo, mid, hi, sig, erf):
    ndim = start.shape[0]
    nlos = start.shape[1]
    inc = np.full(len(shp), 1, dtype=np.int64)
    for i in range(-2, -len(shp)-1, -1):
        inc[i] = inc[i+1]*shp[i+1]

    pmax = np.array(shp)

    out = [None]*nlos
    for i in range(nlos):
        direction = end[:, i]-start[:, i]
        dirx = np.where(direction == 0., 1e-12, direction)
        d0 = np.where(direction == 0., ((start[:, i] > 0)-0.5)*1e12,
                      -start[:, i]/dirx)
        d1 = np.where(direction == 0., ((start[:, i] < pmax)-0.5)*-1e12,
                      (pmax-start[:, i])/dirx)
        (dmin, dmax) = (np.minimum(d0, d1), np.maximum(d0, d1))
        dmin = dmin.max()
        dmax = dmax.min()
        dmin = np.maximum(0., dmin)
        dmax = np.minimum(1., dmax)
        dmax = np.maximum(dmin, dmax)
        # hack: move away from potential grid crossings
        dmin += 1e-7
        dmax -= 1e-7
        if dmin >= dmax:  # no intersection
            out[i] = (np.full(0, 0, dtype=np.int64), np.full(0, 0.))
            continue
        # determine coordinates of first cell crossing
        c_first = np.ceil(start[:, i]+direction*dmin)
        c_first = np.where(direction > 0., c_first, c_first-1.)
        c_first = (c_first-start[:, i])/dirx
        pos1 = np.asarray((start[:, i]+dmin*direction), dtype=np.int64)
        pos1 = np.sum(pos1*inc)
        cdist = np.empty(0, dtype=np.float64)
        add = np.empty(0, dtype=np.int64)
        for j in range(ndim):
            if direction[j] != 0:
                step = inc[j] if direction[j] > 0 else -inc[j]
                tmp = np.arange(start=c_first[j], stop=dmax,
                                step=abs(1./direction[j]))
                cdist = np.append(cdist, tmp)
                add = np.append(add, np.full(len(tmp), step, dtype=np.int64))
        idx = np.argsort(cdist)
        cdist = cdist[idx]
        add = add[idx]
        cdist = np.append(np.full(1, dmin), cdist)
        cdist = np.append(cdist, np.full(1, dmax))
        corfac = np.linalg.norm(direction*dist)
        cdist *= corfac
        wgt = np.diff(cdist)
        mdist = 0.5*(cdist[:-1]+cdist[1:])
        wgt = apply_erf(wgt, mdist, lo[i], mid[i], hi[i], sig[i], erf)
        add = np.append(pos1, add)
        add = np.cumsum(add)
        out[i] = (add, wgt)
    return out


def apply_erf(wgt, dist, lo, mid, hi, sig, erf):
    wgt = wgt.copy()
    mask = dist > hi
    wgt[mask] = 0.
    mask = (dist > lo) & (dist <= hi)
    wgt[mask] *= erf((-1/dist[mask]+1/mid)/sig)
    return wgt


class LOSResponse(LinearOperator):
    """Line-of-sight response operator

    This operator transforms from a single RGSpace to an UnstructuredDomain
    with as many entries as there were lines of sight passed to the
    constructor. Adjoint application is also provided.

    Parameters
    ----------
    domain : RGSpace or DomainTuple
        The operator's input domain. This must be a single RGSpace.
    starts, ends : numpy.ndarray(float) with two dimensions
        Arrays containing the start and end points of the individual lines
        of sight. The first dimension must have as many entries as `domain`
        has dimensions. The second dimensions must be identical for both arrays
        and indicated the total number of lines of sight.
    sigmas: numpy.ndarray(float) (optional)
        If this is not None, the inverse of the lengths of the LOSs are assumed
        to be Gaussian distributed with these sigmas. The start point will
        remain the same, but the endpoint is assumed to be unknown.
        This is a typical statistical model for astrophysical parallaxes.
        The LOS response then returns the expected integral
        over the input given that the length of the LOS is unknown and
        therefore the result is averaged over different endpoints.
        Default: None.
    truncation: float (optional)
        Use only if the sigmas keyword argument is used!
        This truncates the probability of the endpoint lying more sigmas away
        than the truncation. Used to speed up computation and to avoid negative
        distances. It should hold that `1./(1./length-sigma*truncation)>0`
        for all lengths of the LOSs and all corresponding sigma of sigmas.
        If unsure, leave blank.
        Default: 3.

    Notes
    -----
    `starts, `ends`, `sigmas`, and `truncation` have to be identical on
    every calling MPI task (i.e. the full LOS information has to be provided on
    every task).
    """

    def __init__(self, domain, starts, ends, sigmas=None, truncation=3.):
        self._domain = DomainTuple.make(domain)
        self._capability = self.TIMES | self.ADJOINT_TIMES

        if ((not isinstance(self.domain[0], RGSpace)) or
                (len(self._domain) != 1)):
            raise TypeError("The domain must be exactly one RGSpace instance.")

        ndim = len(self.domain[0].shape)
        starts = np.array(starts)
        nlos = starts.shape[1]
        ends = np.array(ends)
        if sigmas is None:
            sigmas = np.zeros(nlos, dtype=np.float32)
        sigmas = np.array(sigmas)
        if starts.shape[0] != ndim:
            raise TypeError("dimension mismatch")
        if nlos != sigmas.shape[0]:
            raise TypeError("dimension mismatch")
        if starts.shape != ends.shape:
            raise TypeError("dimension mismatch")

        diffs = ends-starts
        difflen = np.linalg.norm(diffs, axis=0)
        diffs /= difflen
        real_distances = 1./(1./difflen - truncation*sigmas)
        if np.any(real_distances < 0):
            raise ValueError("parallax error truncation to high: "
                             "getting negative distances")
        real_ends = starts + diffs*real_distances
        dist = np.array(self.domain[0].distances).reshape((-1, 1))
        pixel_starts = starts/dist + 0.5
        pixel_ends = real_ends/dist + 0.5

        w_i = _comp_traverse(pixel_starts,
                             pixel_ends,
                             self.domain[0].shape,
                             np.array(self.domain[0].distances),
                             1./(1./difflen+truncation*sigmas),
                             difflen,
                             1./(1./difflen-truncation*sigmas),
                             sigmas,
                             _gaussian_sf)

        boxsz = 16
        nlos = len(w_i)
        npix = np.prod(self.domain[0].shape)
        ntot = 0
        for i in w_i:
            ntot += len(i[1])
        pri = np.empty(ntot, dtype=np.float64)
        ilos = np.empty(ntot, dtype=np.int32)
        iarr = np.empty(ntot, dtype=np.int32)
        xwgt = np.empty(ntot, dtype=np.float32)
        ofs = 0
        cnt = 0
        for i in w_i:
            nval = len(i[1])
            ilos[ofs:ofs+nval] = cnt
            iarr[ofs:ofs+nval] = i[0]
            xwgt[ofs:ofs+nval] = i[1]
            fullidx = np.unravel_index(i[0], self.domain[0].shape)
            tmp = np.zeros(nval, dtype=np.float64)
            fct = 1.
            for j in range(ndim):
                tmp += (fullidx[j]//boxsz)*fct
                fct *= self.domain[0].shape[j]
            tmp += cnt/float(nlos)
            tmp += iarr[ofs:ofs+nval]/(float(nlos)*float(npix))
            pri[ofs:ofs+nval] = tmp
            ofs += nval
            cnt += 1
        xtmp = np.argsort(pri)
        ilos = ilos[xtmp]
        iarr = iarr[xtmp]
        xwgt = xwgt[xtmp]
        self._smat = aslinearoperator(
            coo_matrix((xwgt, (ilos, iarr)),
                       shape=(nlos, np.prod(self.domain[0].shape))))

        self._target = DomainTuple.make(UnstructuredDomain(nlos))

    def apply(self, x, mode):
        self._check_input(x, mode)
        if mode == self.TIMES:
            result_arr = self._smat.matvec(x.val.reshape(-1))
            return Field(self._target, result_arr)
        input_data = x.val.reshape(-1)
        res = self._smat.rmatvec(input_data).reshape(self.domain[0].shape)
        return Field(self._domain, res)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import sys
from functools import partial

import pytest

pytest.importorskip("jax")
pytest.importorskip("healpy")

import jax
from jax import random
import jax.numpy as jnp
from jax.tree_util import Partial
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

import nifty8.re as jft

pmp = pytest.mark.parametrize

jax.config.update("jax_enable_x64", True)


def lst2fixt(lst):
    @pytest.fixture(params=lst)
    def fixt(request):
        return request.param

    return fixt


def matern_kernel(distance, scale, cutoff, dof):
    from jax.scipy.special import gammaln
    from scipy.special import kv

    reg_dist = jnp.sqrt(2 * dof) * distance / cutoff
    return scale**2 * 2**(1 - dof) / jnp.exp(
        gammaln(dof)
    ) * (reg_dist)**dof * kv(dof, reg_dist)


scale, cutoff, dof = 1., 2., 3 / 2

x = jnp.logspace(-6, 11, base=jnp.e, num=int(1e+5))
y = matern_kernel(x, scale, cutoff, dof)
y = jnp.nan_to_num(y, nan=0.)
kernel = Partial(jnp.interp, xp=x, fp=y)
inv_kernel = Partial(jnp.interp, xp=y, fp=x)

cc_0 = jft.HEALPixChart(
    min_shape=(
        12 * 2**2,
        6,
    ),
    nonhp_rg2cart=lambda x: (jnp.exp(0.3 * x[0] - 0.3), ),
    nonhp_cart2rg=lambda x: ((jnp.log(x[0]) + 0.3) / 0.3, ),
    depth=1,
    _coarse_size=3,
    _fine_size=2,
)
cc_1 = jft.HEALPixChart(
    min_shape=(
        12 * 4**2,
        6,
    ),
    nonhp_rg2cart=lambda x: (jnp.exp(0.75 * x[0] - 1.), ),
    nonhp_cart2rg=lambda x: ((jnp.log(x[0]) + 1.) / 0.75, ),
    depth=1,
    _coarse_size=3,
    _fine_size=2,
)
cc = lst2fixt((cc_0, cc_1))


@pmp(
    "atol,rtol,which", (
        (1e-2 * scale, 5e-2 * scale, "cov"),
        (1e-3 * scale, 5e-3 * scale, "cov"),
        (1e-4 * scale, 5e-2 * scale, "dist")
    )
)
def test_healpix_refinement_matrices_uniquifying(
    cc, atol, rtol, which, kernel=kernel
):
    seed = 42
    aassert = partial(
        assert_allclose,
        atol=1e-11 * scale,
        rtol=1e-10 * scale,
        equal_nan=False
    )

    rf = jft.RefinementHPField(cc, kernel=kernel)
    rfm = rf.matrices()
    mbs = 12 * cc.nside**2 - 1
    rfm_u = rf.matrices(atol=atol, rtol=rtol, which=which, mat_buffer_size=mbs)
    assert rfm_u.index_map[-1].size == cc.shape_at(cc.depth - 1)[0]
    # Assert that we are actually using fewer matrices
    assert np.unique(rfm_u.index_map[-1]).size < rfm_u.index_map[-1].size
    msg = f"{np.unique(rfm_u.index_map[-1]).size} !< {cc.shape_at(cc.depth - 1)[0]}"
    print(msg, file=sys.stderr)
    assert_array_equal(rfm_u.cov_sqrt0, rfm.cov_sqrt0)
    for lvl in range(cc.depth):
        naive_f, naive_p = rfm.filter[lvl], rfm.propagator_sqrt[lvl]
        uniq_f = rfm_u.filter[lvl][rfm_u.index_map[lvl]]
        uniq_p = rfm_u.propagator_sqrt[lvl][rfm_u.index_map[lvl]]
        aassert(uniq_f, naive_f, rtol=1e-7)
        aassert(uniq_p, naive_p, rtol=1e-7)

    key = random.PRNGKey(seed)
    key, sk = random.split(key)
    exc = jft.random_like(sk, rf.domain)
    f = rf(exc, kernel=rfm)
    f_u = rf(exc, kernel=rfm_u)
    aassert(f_u, f)


@pmp("atol,rtol,which", ((1e-2 * scale, 5e-2 * scale, "cov"), (None, ) * 3))
def test_healpix_refinement(cc, atol, rtol, which, kernel=kernel):
    cov_sqrt = jft.refine.healpix_refine.cov_sqrt(cc, kernel, cc.depth)
    cov = cov_sqrt @ cov_sqrt.T

    cf = jft.RefinementHPField(cc, kernel)
    mbs = 12 * cc.nside**2 - 1
    rfm = cf.matrices(atol=atol, rtol=rtol, which=which, mat_buffer_size=mbs)
    if which is not None:
        assert rfm.index_map is not None
        # Assert that we are actually using fewer matrices
        assert np.unique(rfm.index_map[-1]).size < rfm.index_map[-1].size
    cf_wm = jft.Model(partial(cf, kernel=rfm), domain=cf.domain)
    cov_refinement = jft.refine.util.refinement_covariance(cf_wm, kernel)

    assert_allclose(
        cov_refinement,
        cov,
        atol=0.2 * scale,
        rtol=0.1 * scale,
        equal_nan=False
    )


if __name__ == "__main__":
    test_healpix_refinement_matrices_uniquifying(
        cc_0, atol=1e-2, rtol=1e-5, which="cov", kernel=kernel
    )
    test_healpix_refinement(
        cc_0, atol=1e-2, rtol=5e-2, which="cov", kernel=kernel
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020 Max-Planck-Society
# Author: Rouven Lemmerz
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from copy import deepcopy

import nifty8 as ift
import numpy as np
from numpy.testing import assert_allclose


class CountingOp(ift.Operator):
    # FIXME: Not a LinearOperator since ChainOps not supported yet
    def __init__(self, domain):
        self._domain = self._target = ift.sugar.makeDomain(domain)
        self._count = 0

    def apply(self, x):
        self._count += 1
        return x

    @property
    def count(self):
        return self._count


def test_operator_tree_optimiser():
    dom = ift.RGSpace(10, harmonic=True)
    cop1 = CountingOp(dom)
    op1 = (ift.UniformOperator(dom, -1, 2)@cop1).ducktape('a')
    cop2 = CountingOp(dom)
    op2 = ift.FieldZeroPadder(dom, (11,))@cop2
    cop3 = CountingOp(op2.target)
    op3 = ift.ScalingOperator(op2.target, 3)@cop3
    cop4 = CountingOp(op2.target)
    op4 = ift.ScalingOperator(op2.target, 1.5) @ cop4
    op1 = op1 * op1
    # test layering in between two levels
    op = op3@op2@op1 + op2@op1 + op3@op2@op1 + op2@op1
    op = op + op
    op = op4@(op4@op + op4@op)
    fld = ift.from_random(op.domain, 'normal', np.float64)
    op_orig = deepcopy(op)
    op = ift.operator_tree_optimiser._optimise_operator(op)
    assert_allclose(op(fld).val, op_orig(fld).val, rtol=np.finfo(np.float64).eps)
    ift.myassert(1 == ((cop4.count-1) * cop3.count * cop2.count * cop1.count))
    # test testing
    ift.optimise_operator(op_orig)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np


class RandomStateIsSane:
    def __init__(self):
        self._n_states = None

    def __enter__(self):
        self._n_states = len(ift.random._rng)
        assert self._n_states == len(ift.random._sseq)

    def __exit__(self, exc_type, exc_value, tb):
        assert self._n_states == len(ift.random._rng)
        assert self._n_states == len(ift.random._sseq)
        return exc_type is None


def test_rand1():
    with RandomStateIsSane():
        with ift.random.Context(31):
            a = ift.random.current_rng().integers(0,1000000000)
        with ift.random.Context(31):
            b = ift.random.current_rng().integers(0,1000000000)
    np.testing.assert_equal(a,b)


def test_rand2():
    with RandomStateIsSane():
        sseq = ift.random.spawn_sseq(10)
        with ift.random.Context(sseq[2]):
            a = ift.random.current_rng().integers(0,1000000000)
        with ift.random.Context(sseq[2]):
            b = ift.random.current_rng().integers(0,1000000000)
    np.testing.assert_equal(a,b)


def test_rand3():
    with RandomStateIsSane():
        with ift.random.Context(31):
            sseq = ift.random.spawn_sseq(10)
            with ift.random.Context(sseq[2]):
                a = ift.random.current_rng().integers(0,1000000000)
        with ift.random.Context(31):
            sseq = ift.random.spawn_sseq(1)
            sseq = ift.random.spawn_sseq(1)
            sseq = ift.random.spawn_sseq(1)
            with ift.random.Context(sseq[0]):
                b = ift.random.current_rng().integers(0,1000000000)
    np.testing.assert_equal(a,b)


def test_rand4():
    with ift.random.Context(31):
        a = ift.random.current_rng().integers(0,1000000000)
        with ift.random.Context(31):
            b = ift.random.current_rng().integers(0,1000000000)
    np.testing.assert_equal(a,b)


def test_rand5():
    with RandomStateIsSane():
        ift.random.push_sseq_from_seed(31)
        a = ift.random.current_rng().integers(0,1000000000)
        ift.random.push_sseq_from_seed(31)
        b = ift.random.current_rng().integers(0,1000000000)
        c = ift.random.current_rng().integers(0,1000000000)
        ift.random.pop_sseq()
        d = ift.random.current_rng().integers(0,1000000000)
        ift.random.pop_sseq()
    np.testing.assert_equal(a,b)
    np.testing.assert_equal(c,d)


def test_rand5b():
    with RandomStateIsSane():
        with ift.random.Context(31):
            a = ift.random.current_rng().integers(0,1000000000)
            with ift.random.Context(31):
                b = ift.random.current_rng().integers(0,1000000000)
                c = ift.random.current_rng().integers(0,1000000000)
            d = ift.random.current_rng().integers(0,1000000000)
    np.testing.assert_equal(a,b)
    np.testing.assert_equal(c,d)


def test_rand6():
    with RandomStateIsSane():
        ift.random.push_sseq_from_seed(31)
        state = ift.random.getState()
        a = ift.random.current_rng().integers(0,1000000000)
        ift.random.setState(state)
        b = ift.random.current_rng().integers(0,1000000000)
        np.testing.assert_equal(a,b)
        ift.random.pop_sseq()


def test_rand6b():
    with RandomStateIsSane():
        with ift.random.Context(31):
            state = ift.random.getState()
            a = ift.random.current_rng().integers(0,1000000000)
            ift.random.setState(state)
            b = ift.random.current_rng().integers(0,1000000000)
            np.testing.assert_equal(a,b)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from .common import setup_function, teardown_function

pmp = pytest.mark.parametrize
IC = ift.GradientNormController(tol_abs_gradnorm=1e-5, iteration_limit=1000)
SIC = ift.StochasticAbsDeltaEnergyController(1e-8, iteration_limit=100000)
spaces = [ift.RGSpace([1024], distances=0.123), ift.HPSpace(32)]

minimizers = [
    'ift.VL_BFGS(IC)',
    'ift.NonlinearCG(IC, "Polak-Ribiere")',
    # 'ift.NonlinearCG(IC, "Hestenes-Stiefel"),
    'ift.NonlinearCG(IC, "Fletcher-Reeves")',
    'ift.NonlinearCG(IC, "5.49")',
    'ift.L_BFGS_B(ftol=1e-10, gtol=1e-5, maxiter=1000)',
    'ift.L_BFGS(IC)',
    'ift.NewtonCG(IC)'
]

newton_minimizers = ['ift.RelaxedNewton(IC)']
quadratic_only_minimizers = [
    'ift.ConjugateGradient(IC)',
    'ift.minimization.scipy_minimizer._ScipyCG(tol=1e-5, maxiter=300)'
]
slow_minimizers = ['ift.SteepestDescent(IC)', 'ift.ADVIOptimizer(SIC, resample=False)']


@pmp('minimizer', minimizers + newton_minimizers + quadratic_only_minimizers +
     slow_minimizers)
@pmp('space', spaces)
def test_quadratic_minimization(minimizer, space):
    with ift.random.Context(98765):
        starting_point = ift.Field.from_random(domain=space, random_type='normal') * 10
        covariance_diagonal = ift.Field.from_random(domain=space, random_type='uniform') + 0.5
        covariance = ift.DiagonalOperator(covariance_diagonal)
        required_result = ift.full(space, 1.)

        try:
            minimizer = eval(minimizer)
            energy = ift.QuadraticEnergy(
                A=covariance, b=required_result, position=starting_point)

            (energy, convergence) = minimizer(energy)
        except NotImplementedError:
            pytest.skip()

        assert_equal(convergence, IC.CONVERGED)
        assert_allclose(
            energy.position.val,
            1./covariance_diagonal.val,
            rtol=1e-3,
            atol=1e-3)


@pmp('space', spaces)
def test_WF_curvature(space):
    required_result = ift.full(space, 1.)

    s = ift.Field.from_random(domain=space, random_type='uniform') + 0.5
    S = ift.makeOp(s, sampling_dtype=np.float64)
    r = ift.Field.from_random(domain=space, random_type='uniform')
    R = ift.DiagonalOperator(r)
    n = ift.Field.from_random(domain=space, random_type='uniform') + 0.5
    N = ift.makeOp(n, sampling_dtype=np.float64)
    all_diag = 1./s + r**2/n
    curv = ift.WienerFilterCurvature(R, N, S, iteration_controller=IC,
                                     iteration_controller_sampling=IC)
    for b in [False, True]:
        S.draw_sample(b)
        N.draw_sample(b)

    m = curv.inverse(required_result)
    assert_allclose(
        m.val,
        1./all_diag.val,
        rtol=1e-3,
        atol=1e-3)
    curv.draw_sample()
    curv.draw_sample(from_inverse=True)

    if len(space.shape) == 1:
        R = ift.ValueInserter(space, [0])
        n = ift.from_random(R.domain, 'uniform') + 0.5
        N = ift.makeOp(n, sampling_dtype=np.float64)
        all_diag = 1./s + R(1/n)
        curv = ift.WienerFilterCurvature(R.adjoint, N, S,
                                         iteration_controller=IC,
                                         iteration_controller_sampling=IC)
        m = curv.inverse(required_result)
        assert_allclose(
            m.val,
            1./all_diag.val,
            rtol=1e-3,
            atol=1e-3)
        curv.draw_sample()
        curv.draw_sample(from_inverse=True)


@pmp('minimizer', minimizers + newton_minimizers)
def test_rosenbrock(minimizer):
    try:
        from scipy.optimize import rosen, rosen_der, rosen_hess_prod
    except ImportError:
        pytest.skip()
    space = ift.DomainTuple.make(ift.UnstructuredDomain((2,)))
    starting_point = ift.Field.from_random(domain=space, random_type='normal') * 10

    class RBEnergy(ift.Energy):
        def __init__(self, position):
            super(RBEnergy, self).__init__(position)

        @property
        def value(self):
            return rosen(self._position.val_rw())

        @property
        def gradient(self):
            inp = self._position.val_rw()
            out = ift.Field(space, rosen_der(inp))
            return out

        @property
        def metric(self):
            class RBCurv(ift.EndomorphicOperator):
                def __init__(self, loc):
                    self._loc = loc.val_rw()
                    self._capability = self.TIMES
                    self._domain = space

                def apply(self, x, mode):
                    self._check_input(x, mode)
                    inp = x.val_rw()
                    out = ift.Field(
                        space, rosen_hess_prod(self._loc.copy(), inp))
                    return out

            t1 = ift.GradientNormController(
                tol_abs_gradnorm=1e-5, iteration_limit=1000)
            return ift.InversionEnabler(RBCurv(self._position), t1)

        def apply_metric(self, x):
            inp = x.val_rw()
            pos = self._position.val_rw()
            return ift.Field(space, rosen_hess_prod(pos, inp))

    try:
        minimizer = eval(minimizer)
        energy = RBEnergy(position=starting_point)

        (energy, convergence) = minimizer(energy)
    except NotImplementedError:
        pytest.skip()

    assert_equal(convergence, IC.CONVERGED)
    assert_allclose(energy.position.val, 1., rtol=1e-3, atol=1e-3)


@pmp('minimizer', minimizers + slow_minimizers)
def test_gauss(minimizer):
    space = ift.UnstructuredDomain((1,))
    starting_point = ift.Field.full(space, 3.)

    class ExpEnergy(ift.Energy):
        def __init__(self, position):
            super(ExpEnergy, self).__init__(position)

        @property
        def value(self):
            x = self.position.val[0]
            return -np.exp(-(x**2))

        @property
        def gradient(self):
            x = self.position.val[0]
            return ift.Field.full(self.position.domain, 2*x*np.exp(-(x**2)))

        def apply_metric(self, x):
            p = self.position.val[0]
            v = (2 - 4*p*p)*np.exp(-p**2)
            return ift.DiagonalOperator(
                ift.Field.full(self.position.domain, v))(x)

    try:
        minimizer = eval(minimizer)
        energy = ExpEnergy(position=starting_point)

        (energy, convergence) = minimizer(energy)
    except NotImplementedError:
        pytest.skip()

    assert_equal(convergence, IC.CONVERGED)
    assert_allclose(energy.position.val, 0., atol=1e-3)


@pmp('minimizer', minimizers + newton_minimizers + slow_minimizers)
def test_cosh(minimizer):
    space = ift.UnstructuredDomain((1,))
    starting_point = ift.Field.full(space, 3.)

    class CoshEnergy(ift.Energy):
        def __init__(self, position):
            super(CoshEnergy, self).__init__(position)

        @property
        def value(self):
            x = self.position.val[0]
            return np.cosh(x)

        @property
        def gradient(self):
            x = self.position.val[0]
            return ift.Field.full(self.position.domain, np.sinh(x))

        @property
        def metric(self):
            x = self.position.val[0]
            v = np.cosh(x)
            return ift.DiagonalOperator(
                ift.Field.full(self.position.domain, v))

        def apply_metric(self, x):
            p = self.position.val[0]
            v = np.cosh(p)
            return ift.DiagonalOperator(
                ift.Field.full(self.position.domain, v))(x)

    try:
        minimizer = eval(minimizer)
        energy = CoshEnergy(position=starting_point)

        (energy, convergence) = minimizer(energy)
    except NotImplementedError:
        pytest.skip()

    assert_equal(convergence, IC.CONVERGED)
    assert_allclose(energy.position.val, 0., atol=1e-3)


@pmp('space', spaces)
def test_scipy_respect_bounds(space):
    starting_point = ift.Field.from_random(domain=space, random_type='normal') * 10
    covariance_diagonal = ift.Field.from_random(domain=space, random_type='uniform') + 0.5
    covariance = ift.DiagonalOperator(covariance_diagonal)
    required_result = ift.full(space, 1.)
    left_bound = 2.
    best_possible_result = ift.full(space, left_bound)

    minimizer = ift.L_BFGS_B(ftol=1e-10, gtol=1e-5, maxiter=1000, bounds=(left_bound, 10))

    energy = ift.QuadraticEnergy(
        A=covariance, b=required_result, position=starting_point)

    (energy, convergence) = minimizer(energy)

    assert_allclose(energy.position.val, best_possible_result.val)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Max-Planck-Society, Philipp Arras

from .endomorphic_operator import EndomorphicOperator
from .operator import Operator


class CountingOperator(Operator):
    def __init__(self, domain):
        from ..sugar import makeDomain

        self._domain = self._target = makeDomain(domain)
        self._count_apply = 0
        self._count_apply_lin = 0
        self._derivative = _JacCountingOperator(self._domain)

    def apply(self, x):
        from ..sugar import is_linearization

        self._check_input(x)
        if is_linearization(x):
            self._count_apply_lin += 1
            x = x.new(x.val, self._derivative)
        else:
            self._count_apply += 1
        return x

    @property
    def count_apply(self):
        return self._count_apply

    @property
    def count_apply_lin(self):
        return self._count_apply_lin

    @property
    def count_jac(self):
        return self._derivative._count_times

    @property
    def count_jac_adj(self):
        return self._derivative._count_adjoint_times

    def __repr__(self):
        return f"CountingOperator({self._domain.__repr__()})"

    def report(self):
        s = [f"* apply: \t\t{self.count_apply:>7}",
             f"* apply Linearization: \t{self.count_apply_lin:>7}",
             f"* Jacobian: \t\t{self.count_jac:>7}",
             f"* Adjoint Jacobian: \t{self.count_jac_adj:>7}"]
        return "\n".join(s)

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from .simplify_for_const import InsertionOperator
        return None, self @ InsertionOperator(self.domain, c_inp)


class _JacCountingOperator(EndomorphicOperator):
    def __init__(self, domain):
        from ..sugar import makeDomain

        self._domain = makeDomain(domain)
        self._capability = self.TIMES | self.ADJOINT_TIMES
        self._count_times = 0
        self._count_adjoint_times = 0

    def apply(self, x, mode):
        self._check_input(x, mode)
        if mode == self.TIMES:
            self._count_times += 1
        else:
            self._count_adjoint_times += 1
        return x



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from .energy import Energy


class QuadraticEnergy(Energy):
    """The Energy for a quadratic form.
    The most important aspect of this energy is that its metric must be
    position-independent.
    """

    def __init__(self, position, A, b, _grad=None):
        super(QuadraticEnergy, self).__init__(position=position)
        self._A = A
        self._b = b
        if _grad is not None:
            self._grad = _grad
            Ax = _grad if b is None else _grad + b
        else:
            Ax = self._A(self._position)
            self._grad = Ax if b is None else Ax - b
        self._value = 0.5*self._position.s_vdot(Ax).real
        if b is not None:
            self._value -= b.s_vdot(self._position).real

    def at(self, position):
        return QuadraticEnergy(position, self._A, self._b)

    def at_with_grad(self, position, grad):
        """Specialized version of `at`, taking also a gradient.

        This custom method is meant for use within :class:ConjugateGradient`
        minimizers, which already have the gradient available. It saves time
        by not recomputing it.

        Parameters
        ----------
        position : :class:`nifty8.field.Field`
            Location in parameter space for the new Energy object.
        grad : :class:`nifty8.field.Field`
            Energy gradient at the new position.

        Returns
        -------
        Energy
            Energy object at new position.
        """
        return QuadraticEnergy(position, self._A, self._b, grad)

    @property
    def value(self):
        return self._value

    @property
    def gradient(self):
        return self._grad

    @property
    def metric(self):
        return self._A

    def apply_metric(self, x):
        return self._A(x)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from functools import partial

import numpy as np

from ..domain_tuple import DomainTuple
from ..domains.unstructured_domain import UnstructuredDomain
from ..field import Field
from ..multi_domain import MultiDomain
from ..multi_field import MultiField
from ..utilities import check_object_identity
from .endomorphic_operator import EndomorphicOperator
from .linear_operator import LinearOperator


class VdotOperator(LinearOperator):
    """Operator computing the scalar product of its input with a given Field.

    Parameters
    ----------
    field : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
        The field used to build the scalar product with the operator input
    """
    def __init__(self, field):
        self._field = field
        self._domain = field.domain
        self._target = DomainTuple.scalar_domain()
        self._capability = self.TIMES | self.ADJOINT_TIMES

        try:
            from ..re import vdot

            self._jax_expr = partial(vdot, field.val)
        except ImportError:
            self._jax_expr = None

    def apply(self, x, mode):
        self._check_mode(mode)
        if mode == self.TIMES:
            return self._field.vdot(x)
        return self._field*x.val[()]


class ConjugationOperator(EndomorphicOperator):
    """Operator computing the complex conjugate of its input.

    Parameters
    ----------
    domain: Domain, tuple of domains or DomainTuple
        domain of the input field

    """
    def __init__(self, domain):
        self._domain = DomainTuple.make(domain)
        self._capability = self._all_ops

        try:
            from jax import numpy as jnp
            from jax.tree_util import tree_map

            self._jax_expr = partial(tree_map, jnp.conjugate)
        except ImportError:
            self._jax_expr = None

    def apply(self, x, mode):
        self._check_input(x, mode)
        return x.conjugate()


class WeightApplier(EndomorphicOperator):
    """Operator multiplying its input by a given power of dvol.

    Parameters
    ----------
    domain: Domain, tuple of domains or DomainTuple
        domain of the input field
    spaces: list or tuple of int
        indices of subdomains for which the weights shall be applied
    power: int
        the power of to be used for the volume factors

    """
    def __init__(self, domain, spaces, power):
        from .. import utilities
        self._domain = DomainTuple.make(domain)
        if spaces is None:
            self._spaces = None
        else:
            self._spaces = utilities.parse_spaces(spaces, len(self._domain))
        self._power = int(power)
        self._capability = self._all_ops

    def apply(self, x, mode):
        self._check_input(x, mode)
        power = self._power if (mode & 3) else -self._power
        return x.weight(power, spaces=self._spaces)


class Realizer(EndomorphicOperator):
    """Operator returning the real component of its input.

    Parameters
    ----------
    domain: Domain, tuple of domains, DomainTuple or MultiDomain
        domain of the input field

    """
    def __init__(self, domain):
        from ..sugar import makeDomain

        self._domain = makeDomain(domain)
        self._capability = self.TIMES | self.ADJOINT_TIMES

        try:
            from jax import numpy as jnp
            from jax.tree_util import tree_map

            self._jax_expr = partial(tree_map, jnp.real)
        except ImportError:
            self._jax_expr = None

    def apply(self, x, mode):
        self._check_input(x, mode)
        return x.real


class Imaginizer(EndomorphicOperator):
    """Operator returning the imaginary component of its input.

    Parameters
    ----------
    domain: Domain, tuple of domains, DomainTuple or MultiDomain
        domain of the input field

    """
    def __init__(self, domain):
        from ..sugar import makeDomain

        self._domain = makeDomain(domain)
        self._capability = self.TIMES | self.ADJOINT_TIMES

        try:
            from jax import numpy as jnp
            from jax.tree_util import tree_map

            self._jax_expr = partial(tree_map, jnp.imag)
        except ImportError:
            self._jax_expr = None

    def apply(self, x, mode):
        self._check_input(x, mode)
        if mode == self.TIMES:
            if not np.issubdtype(x.dtype, np.complexfloating):
                raise ValueError
            return x.imag
        if x.dtype not in (np.float64, np.float32):
            raise ValueError
        return 1j*x


class FieldAdapter(LinearOperator):
    """Operator for conversion between Fields and MultiFields.

    Parameters
    ----------
    tgt : Domain, tuple of domains, DomainTuple, dict or MultiDomain:
        If this is a Domain, tuple of Domain or DomainTuple, this will be the
        operator's target, and its domain will be a MultiDomain consisting of
        its domain with the supplied `name`
        If this is a dict or MultiDomain, everything except for `name` will
        be stripped out of it, and the result will be the operator's target.
        Its domain will then be the DomainTuple corresponding to the single
        entry in the operator's domain.

    name : String
        The relevant key of the MultiDomain.
    """

    def __init__(self, target, name):
        from ..sugar import makeDomain
        tmp = makeDomain(target)
        if isinstance(tmp, DomainTuple):
            self._target = tmp
            self._domain = MultiDomain.make({name: tmp})
        else:
            self._domain = tmp[name]
            self._target = MultiDomain.make({name: tmp[name]})
        self._capability = self.TIMES | self.ADJOINT_TIMES

        try:
            from .. import re as jft

            def wrap(x):
                return jft.Vector({name: x})

            def unwrap(x):
                return x[name]

            if isinstance(tmp, DomainTuple):
                self._jax_expr = unwrap
            else:
                self._jax_expr = wrap
        except ImportError:
            self._jax_expr = None

    def apply(self, x, mode):
        self._check_input(x, mode)
        if isinstance(x, MultiField):
            return x.values()[0]
        else:
            return MultiField(self._tgt(mode), (x,))

    def __repr__(self):
        dom = self.domain.keys() if isinstance(self.domain, MultiDomain) else '()'
        tgt = self.target.keys() if isinstance(self.target, MultiDomain) else '()'
        return f'{tgt} <- {dom}'


class _SlowFieldAdapter(LinearOperator):
    """Operator for conversion between Fields and MultiFields.
    The operator is built so that the MultiDomain is always the target.
    Its domain is `tgt[name]`

    Parameters
    ----------
    dom : dict or MultiDomain:
        the operator's dom

    name : String
        The relevant key of the MultiDomain.
    """

    def __init__(self, domain, name):
        from ..sugar import makeDomain
        tmp = makeDomain(domain)
        if not isinstance(tmp, MultiDomain):
            raise TypeError("MultiDomain expected")
        self._name = str(name)
        self._domain = tmp
        self._target = tmp[name]
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        if isinstance(x, MultiField):
            return x[self._name]
        return MultiField.from_dict({self._name: x}, domain=self._tgt(mode))

    def __repr__(self):
        return '_SlowFieldAdapter'


def ducktape(left, right, name):
    """Convenience function creating an operator that converts between a
    DomainTuple and a MultiDomain.

    Parameters
    ----------
    left : None, Operator, or Domainoid
        Something describing the new operator's target domain.
        If `left` is an `Operator`, its domain is used as `left`.

    right : None, Operator, or Domainoid
        Something describing the new operator's input domain.
        If `right` is an `Operator`, its target is used as `right`.

    name : string
        The component of the `MultiDomain` that will be extracted/inserted

    Notes
    -----
    - one of the involved domains must be a `DomainTuple`, the other a
      `MultiDomain`.
    - `left` and `right` must not be both `None`, but one of them can (and
      probably should) be `None`. In this case, the missing information is
      inferred.

    Returns
    -------
    FieldAdapter or _SlowFieldAdapter
        an adapter operator converting between the two (possibly
        partially inferred) domains.
    """
    from ..sugar import makeDomain
    from .operator import Operator
    if isinstance(right, Operator):
        right = right.target
    elif right is not None:
        right = makeDomain(right)
    if isinstance(left, Operator):
        left = left.domain
    elif left is not None:
        left = makeDomain(left)
    if left is None:  # need to infer left from right
        if isinstance(right, MultiDomain):
            left = right[name]
        else:
            left = MultiDomain.make({name: right})
    elif right is None:  # need to infer right from left
        if isinstance(left, MultiDomain):
            right = left[name]
        else:
            right = MultiDomain.make({name: left})
    lmulti = isinstance(left, MultiDomain)
    rmulti = isinstance(right, MultiDomain)
    if lmulti + rmulti != 1:
        raise ValueError("need exactly one MultiDomain")
    if lmulti:
        if len(left) == 1:
            return FieldAdapter(left, name)
        else:
            return _SlowFieldAdapter(left, name).adjoint
    if rmulti:
        if len(right) == 1:
            return FieldAdapter(left, name)
        else:
            return _SlowFieldAdapter(right, name)
    raise ValueError("must not arrive here")


class GeometryRemover(LinearOperator):
    """Operator which transforms between a structured and an unstructured
    domain.

    Parameters
    ----------
    domain: Domain, tuple of Domain, or DomainTuple:
        the full input domain of the operator.
    space: int, optional
        The index of the subdomain on which the operator should act.
        If None, it acts on all spaces.

    Notes
    -----
    The operator will convert every sub-domain of its input domain to an
    UnstructuredDomain with the same shape. No weighting by volume factors
    is carried out.
    """

    def __init__(self, domain, space=None):
        self._domain = DomainTuple.make(domain)
        if space is not None:
            tgt = [dom for dom in self._domain]
            tgt[space] = UnstructuredDomain(self._domain[space].shape)
        else:
            tgt = [UnstructuredDomain(dom.shape) for dom in self._domain]
        self._target = DomainTuple.make(tgt)
        self._capability = self.TIMES | self.ADJOINT_TIMES

        def identity(x):
            return x

        self._jax_expr = identity

    def apply(self, x, mode):
        self._check_input(x, mode)
        return x.cast_domain(self._tgt(mode))


class NullOperator(LinearOperator):
    """Operator corresponding to a matrix of all zeros.

    Parameters
    ----------
    domain : DomainTuple or MultiDomain
        input domain
    target : DomainTuple or MultiDomain
        output domain
    """

    def __init__(self, domain, target):
        from ..sugar import makeDomain
        self._domain = makeDomain(domain)
        self._target = makeDomain(target)
        self._capability = self.TIMES | self.ADJOINT_TIMES

    @staticmethod
    def _nullfield(dom):
        if isinstance(dom, DomainTuple):
            return Field(dom, 0.)
        else:
            return MultiField.full(dom, 0.)

    def apply(self, x, mode):
        self._check_input(x, mode)
        return self._nullfield(self._tgt(mode))

    def __repr__(self):
        dom = self.domain.keys() if isinstance(self.domain, MultiDomain) else '()'
        tgt = self.target.keys() if isinstance(self.target, MultiDomain) else '()'
        return f'{tgt} <- NullOperator <- {dom}'


class PartialExtractor(LinearOperator):
    def __init__(self, domain, target):
        if not isinstance(domain, MultiDomain):
            raise TypeError("MultiDomain expected")
        if not isinstance(target, MultiDomain):
            raise TypeError("MultiDomain expected")
        self._domain = domain
        self._target = target
        for key in self._target.keys():
            check_object_identity(self._domain[key], self._target[key])
        self._capability = self.TIMES | self.ADJOINT_TIMES
        self._compldomain = MultiDomain.make({kk: self._domain[kk]
                                              for kk in self._domain.keys()
                                              if kk not in self._target.keys()})

    def apply(self, x, mode):
        self._check_input(x, mode)
        if mode == self.TIMES:
            return x.extract(self._target)
        res0 = MultiField.from_dict({key: x[key] for key in x.domain.keys()})
        res1 = MultiField.full(self._compldomain, 0.)
        return res0.unite(res1)

    def __repr__(self):
        return f'{self.target.keys()} <- {self.domain.keys()}'


class PrependKey(LinearOperator):
    """Prepend a string to all keys of a MultiDomain.

    Parameters
    ----------
    domain : MultiDomain
    pre : str
    """
    def __init__(self, domain, pre):
        if not isinstance(domain, MultiDomain):
            raise ValueError
        from ..sugar import makeDomain
        self._domain = makeDomain(domain)
        self._pre = str(pre)
        target = {self._pre+k: domain[k] for k in domain.keys()}
        self._target = makeDomain(MultiDomain.make(target))
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        if mode == self.TIMES:
            res = {self._pre+k:x[k] for k in self._domain.keys()}
        else:
            res = {k:x[self._pre+k] for k in self._domain.keys()}
        return MultiField.from_dict(res, domain=self._tgt(mode))


class DomainChangerAndReshaper(LinearOperator):
    """Convert nifty domains into each other and reshape field.

    This is only possible if `domain` and `target` have the same number of pixels.

    Parameters
    ----------
    domain : DomainTuple
        Domain of the operator
    target : DomainTuple
        Target of the operator
    """
    def __init__(self, domain, target):
        from ..sugar import makeDomain

        self._domain = makeDomain(domain)
        self._target = makeDomain(target)
        if self._domain.size != self._target.size:
            s = ["Domain and target do not have the same number of pixels",
                f"Domain: {self._domain.shape}",
                f"Target: {self._target.shape}"]
            raise ValueError("\n".join(s))
        self._capability = self.TIMES | self.ADJOINT_TIMES
        if isinstance(self._domain, MultiDomain) or isinstance(self._target, MultiDomain):
            raise NotImplementedError("MultiDomains are not supported yet")

    def apply(self, x, mode):
        from ..sugar import makeField

        self._check_input(x, mode)
        x = x.val
        tgt = self._tgt(mode)
        return makeField(tgt, x.reshape(tgt.shape))

    def __repr__(self):
        return f"Reshape {self._shapes(self.target)} <- {self._shapes(self.domain)}"

    @staticmethod
    def _shapes(dom):
        return " ".join([str(dd.shape) for dd in dom])


class ExtractAtIndices(LinearOperator):
    """Extract Field values at given indices along a subspace of a DomainTuple,
    and puts them in a field with an unstructured domain for the given subspace.
    Note: This operator supports having the same index several times.
    If this is not the case also the numerically faster `GeometryRemover`
    and `MaskOperator` might be useful.

    Parameters
    ----------
    domain : DomainTuple
        Domain of the operator
    indices: tuple
        Tuple of indices to extract from the field along a given space.
        The length of the tuple needs to equal the number of axes of the space.
        Each entry of the tuple should be a tuple of indices to take along the
        respective axis of the space.
        Example for a 2d space: indices=((0,1,1,0), (3,4,1,5)) will extract
        the pixels (0,3), (1,4), (1,1) and (0,5).
    space: int
        The index of the space in the domain along which to extract values
    """
    def __init__(self, domain, indices, space=0):
        from ..sugar import makeDomain

        self._domain = makeDomain(domain)
        if not isinstance(indices, tuple):
            raise TypeError("indices need to be a tuple")
        if not len(self._domain[space].shape) == len(indices):
            raise ValueError("Shape of indices don't match dimension of space")
        target = [
                UnstructuredDomain(len(indices[0])) if i == space else sp
                for i, sp in enumerate(self._domain)]
        self._target = makeDomain(target)

        inds = [slice(None)] * len(domain.shape)
        dims_of_sps = [len(sp.shape) for sp in self._domain]
        start_ax_sp = int(np.sum(dims_of_sps[:space]))
        stop_ax_sp = start_ax_sp + dims_of_sps[space]
        for i, ax in enumerate(range(start_ax_sp, stop_ax_sp)):
            inds[ax] = indices[i]
        self._inds = tuple(inds)
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        if mode == self.TIMES:
            res = x.val[self._inds]
        else:
            res = np.zeros(self._domain.shape, dtype=x.dtype)
            np.add.at(res, self._inds, x.val)
        return Field.from_raw(self._tgt(mode), res)



# Copyright(C) 2013-2021 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from .forest_math import (
    assert_arithmetics, get_map, has_arithmetics, map_forest, map_forest_mean,
    mean, mean_and_std, random_like, stack, tree_shape, unstack
)
from .pytree_string import PyTreeString, hide_strings
from .vector import Vector
from .vector_math import (
    ShapeWithDtype, all, any, conj, conjugate, dot, matmul, max, min, norm,
    ones_like, result_type, shape, size, sum, vdot, where, zeros_like
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..domain_tuple import DomainTuple
from ..domains.dof_space import DOFSpace
from ..domains.power_space import PowerSpace
from ..field import Field
from ..utilities import infer_space, special_add_at
from .linear_operator import LinearOperator


class DOFDistributor(LinearOperator):
    """Operator which distributes actual degrees of freedom (dof) according to
    some distribution scheme into a higher dimensional space. This distribution
    scheme is defined by the dofdex, a degree of freedom index, which
    associates the entries within the operator's domain to locations in its
    target. This operator's domain is a DOFSpace, which is defined according to
    the distribution scheme.

    Parameters
    ----------
    dofdex: :class:`nifty8.field.Field` of integers
        An integer Field on exactly one Space establishing the association
        between the locations in the operator's target and the underlying
        degrees of freedom in its domain.
        It has to start at 0 and it increases monotonically, no empty bins are
        allowed.
    target: Domain, tuple of Domain, or DomainTuple, optional
        The target of the operator. If not specified, the domain of the dofdex
        is used.
    space: int, optional:
       The index of the sub-domain on which the operator acts.
       Can be omitted if `target` only has one sub-domain.
    """

    def __init__(self, dofdex, target=None, space=None):
        if target is None:
            target = dofdex.domain
        self._target = DomainTuple.make(target)
        space = infer_space(self._target, space)
        partner = self._target[space]
        if not isinstance(dofdex, Field):
            raise TypeError("dofdex must be a Field")
        if not len(dofdex.domain) == 1:
            raise ValueError("dofdex must be defined on exactly one Space")
        if not np.issubdtype(dofdex.dtype, np.integer):
            raise TypeError("dofdex must contain integer numbers")
        if partner != dofdex.domain[0]:
            raise ValueError("incorrect dofdex domain")

        ldat = dofdex.val
        if ldat.size == 0:  # can happen for weird configurations
            nbin = 0
        else:
            nbin = ldat.max()
        nbin = nbin + 1
        if partner.scalar_dvol is not None:
            wgt = np.bincount(dofdex.val.ravel(), minlength=nbin)
            wgt = wgt*partner.scalar_dvol
        else:
            dvol = Field.from_raw(partner, partner.dvol).val
            wgt = np.bincount(dofdex.val.ravel(),
                              minlength=nbin, weights=dvol)
        # The explicit conversion to float64 is necessary because bincount
        # sometimes returns its result as an integer array, even when
        # floating-point weights are present ...
        wgt = wgt.astype(np.float64, copy=False)
        if (wgt == 0).any():
            raise ValueError("empty bins detected")

        self._init2(dofdex.val, space, DOFSpace(wgt))

    def _init2(self, dofdex, space, other_space):
        self._space = space
        dom = list(self._target)
        dom[self._space] = other_space
        self._domain = DomainTuple.make(dom)
        self._capability = self.TIMES | self.ADJOINT_TIMES

        self._dofdex = dofdex.ravel()
        firstaxis = self._target.axes[self._space][0]
        lastaxis = self._target.axes[self._space][-1]
        arrshape = self._target.shape
        presize = np.prod(arrshape[0:firstaxis], dtype=np.int64)
        postsize = np.prod(arrshape[lastaxis+1:], dtype=np.int64)
        self._hshape = (presize, self._domain[self._space].shape[0], postsize)
        self._pshape = (presize, self._dofdex.size, postsize)

    def _adjoint_times(self, x):
        arr = x.val
        arr = arr.reshape(self._pshape)
        oarr = np.zeros(self._hshape, dtype=x.dtype)
        oarr = special_add_at(oarr, 1, self._dofdex, arr)
        oarr = oarr.reshape(self._domain.shape)
        res = Field.from_raw(self._domain, oarr)
        return res

    def _times(self, x):
        arr = x.val
        arr = arr.reshape(self._hshape)
        oarr = np.empty(self._pshape, dtype=x.dtype)
        oarr[()] = arr[(slice(None), self._dofdex, slice(None))]
        return Field(self._target, oarr.reshape(self._target.shape))

    def apply(self, x, mode):
        self._check_input(x, mode)
        return self._times(x) if mode == self.TIMES else self._adjoint_times(x)


class PowerDistributor(DOFDistributor):
    """Operator which transforms between a PowerSpace and a harmonic domain.

    Parameters
    ----------
    target: Domain, tuple of Domain, or DomainTuple
        the total *target* domain of the operator.
    power_space: PowerSpace, optional
        the input sub-domain on which the operator acts.
        If not supplied, a matching PowerSpace with natural binbounds will be
        used.
    space: int, optional:
       The index of the sub-domain on which the operator acts.
       Can be omitted if `target` only has one sub-domain.
    """

    def __init__(self, target, power_space=None, space=None):
        # Initialize domain and target
        self._target = DomainTuple.make(target)
        self._space = infer_space(self._target, space)
        hspace = self._target[self._space]
        if not hspace.harmonic:
            raise ValueError("Operator requires harmonic target space")
        if power_space is None:
            power_space = PowerSpace(hspace)
        else:
            if not isinstance(power_space, PowerSpace):
                raise TypeError("power_space argument must be a PowerSpace")
            if power_space.harmonic_partner != hspace:
                raise ValueError("power_space does not match its partner")

        self._init2(power_space.pindex, self._space, power_space)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose

from .common import setup_function, teardown_function

pmp = pytest.mark.parametrize


def _lin2grad(lin):
    return lin.jac(ift.full(lin.domain, 1.)).val


def jt(lin, check):
    assert_allclose(_lin2grad(lin), check)


def test_special_gradients():
    dom = ift.UnstructuredDomain((1,))
    f = ift.full(dom, 2.4)
    var = ift.Linearization.make_var(f)
    s = f.val

    jt(var.clip(0, 10), np.ones_like(s))
    jt(var.clip(-1, 0), np.zeros_like(s))

    assert_allclose(
        _lin2grad(ift.Linearization.make_var(0*f).ptw("sinc")), np.zeros(s.shape))
    ift.myassert(np.isnan(_lin2grad(ift.Linearization.make_var(0*f).ptw("abs"))))
    assert_allclose(
        _lin2grad(ift.Linearization.make_var(0*f + 10).ptw("abs")),
        np.ones(s.shape))
    assert_allclose(
        _lin2grad(ift.Linearization.make_var(0*f - 10).ptw("abs")),
        -np.ones(s.shape))


@pmp('f', [
    'log', 'exp', 'sqrt', 'sin', 'cos', 'tan', 'sinc', 'sinh', 'cosh', 'tanh',
    'absolute', 'reciprocal', 'sigmoid', 'log10', 'log1p', 'expm1', 'softplus',
    ('power', 2.), ('exponentiate', 1.1)
])
@pmp('cplxpos', [True, False])
@pmp('cplxdir', [True, False])
@pmp('holomorphic', [True, False])
def test_actual_gradients(f, cplxpos, cplxdir, holomorphic):
    if (cplxpos or cplxdir) and f in ['absolute']:
        return
    if holomorphic and f in ['absolute']:
        # These function are not holomorphic
        return
    dom = ift.UnstructuredDomain((1,))
    fld = ift.full(dom, 2.4)
    if cplxpos:
        fld = fld + 0.21j
    eps = 1e-7
    if cplxdir:
        eps *= 1j
    if holomorphic:
        eps *= (1+0.78j)
    var0 = ift.Linearization.make_var(fld)
    var1 = ift.Linearization.make_var(fld + eps)
    if not isinstance(f, tuple):
        f = (f,)
    f0 = var0.ptw(*f).val.val
    f1 = var1.ptw(*f).val.val
    df1 = _lin2grad(var0.ptw(*f))
    df0 = (f1 - f0)/eps
    assert_allclose(df0, df1, rtol=100*np.abs(eps))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from ..multi_domain import MultiDomain
from ..utilities import myassert
from .block_diagonal_operator import BlockDiagonalOperator
from .energy_operators import EnergyOperator, LikelihoodEnergyOperator
from .operator import Operator
from .scaling_operator import ScalingOperator
from .simple_linear_operators import NullOperator


class ConstCollector:
    def __init__(self):
        self._const = None  # MultiField on the part of the MultiDomain that could be constant
        self._nc = set()  # NoConstant - set of keys that we know cannot be constant

    def mult(self, const, fulldom):
        if const is None:
            self._nc |= set(fulldom.keys())
        else:
            from ..multi_field import MultiField
            self._nc |= set(fulldom.keys()) - set(const.keys())
            if self._const is None:
                self._const = MultiField.from_dict(
                    {key: const[key]
                     for key in const.keys() if key not in self._nc})
            else:  # we know that the domains are identical for products
                self._const = MultiField.from_dict(
                    {key: self._const[key]*const[key]
                     for key in const.keys() if key not in self._nc})

    def add(self, const, fulldom):
        if const is None:
            self._nc |= set(fulldom.keys())
        else:
            from ..multi_field import MultiField
            self._nc |= set(fulldom.keys()) - set(const.keys())
            self._const = const if self._const is None else self._const.unite(const)
            self._const = MultiField.from_dict(
                {key: const[key]
                 for key in const.keys() if key not in self._nc})

    @property
    def constfield(self):
        return self._const


class ConstantOperator(Operator):
    def __init__(self, output, domain={}):
        from ..sugar import makeDomain
        self._domain = makeDomain(domain)
        self._target = makeDomain(output.domain)
        self._output = output

    def apply(self, x):
        from .simple_linear_operators import NullOperator
        self._check_input(x)
        if x.jac is not None:
            return x.new(self._output, NullOperator(self._domain, self._target))
        return self._output

    def __repr__(self):
        tgt = self.target.keys() if isinstance(self.target, MultiDomain) else '()'
        return f'{tgt} <- ConstantOperator'


class ConstantEnergyOperator(EnergyOperator):
    def __init__(self, output):
        from ..field import Field
        from ..sugar import makeDomain
        self._domain = makeDomain({})
        if not isinstance(output, Field):
            output = Field.scalar(float(output))
        self._output = output

    def apply(self, x):
        self._check_input(x)
        if x.jac is not None:
            val = self._output
            jac = NullOperator(self._domain, self._target)
            met = NullOperator(self._domain, self._domain) if x.want_metric else None
            return x.new(val, jac, met)
        return self._output


class ConstantLikelihoodEnergyOperator(LikelihoodEnergyOperator):
    def __init__(self, output):
        super(ConstantLikelihoodEnergyOperator, self).__init__(None, None)
        op = ConstantEnergyOperator(output)
        self.apply = op.apply
        self._domain = op._domain


class InsertionOperator(Operator):
    def __init__(self, target, cst_field):
        from ..multi_field import MultiField
        from ..sugar import makeDomain
        if not isinstance(target, MultiDomain):
            raise TypeError
        if not isinstance(cst_field, MultiField):
            raise TypeError
        self._target = MultiDomain.make(target)
        cstdom = cst_field.domain
        vardom = makeDomain({kk: vv for kk, vv in self._target.items()
                             if kk not in cst_field.keys()})
        self._domain = vardom
        self._cst = cst_field
        jac = {kk: ScalingOperator(vv, 1.) for kk, vv in self._domain.items()}
        self._jac = BlockDiagonalOperator(self._domain, jac) + NullOperator(makeDomain({}), cstdom)

    def apply(self, x):
        myassert(len(set(self._cst.keys()) & set(x.domain.keys())) == 0)
        val = x if x.jac is None else x.val
        val = val.unite(self._cst)
        if x.jac is None:
            return val
        return x.new(val, self._jac)

    def __repr__(self):
        from ..utilities import indent
        subs = f'Constant: {self._cst.keys()}\nVariable: {self._domain.keys()}'
        return 'InsertionOperator\n'+indent(subs)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np

Nsamp = 20000
np.random.seed(42)


def _to_array(d):
    if isinstance(d, np.ndarray):
        return d
    assert isinstance(d, dict)
    return np.concatenate(list(d.values()))


def test_GaussianEnergy():
    sp = ift.UnstructuredDomain(Nsamp)
    S = ift.ScalingOperator(sp, 1., complex)
    samp = S.draw_sample()
    real_std = np.std(samp.val.real)
    imag_std = np.std(samp.val.imag)
    np.testing.assert_allclose(real_std, imag_std,
                               atol=5./np.sqrt(Nsamp))
    sp1 = ift.UnstructuredDomain(1)
    mean = ift.full(sp1, 0.)
    real_icov = ift.ScalingOperator(sp1, real_std**(-2))
    imag_icov = ift.ScalingOperator(sp1, imag_std**(-2))
    real_energy = ift.GaussianEnergy(mean, inverse_covariance=real_icov)
    imag_energy = ift.GaussianEnergy(mean, inverse_covariance=imag_icov)
    icov = ift.ScalingOperator(sp1, 1.)
    complex_energy = ift.GaussianEnergy(mean+0.j, inverse_covariance=icov)
    for i in range(min(10, Nsamp)):
        fld = ift.full(sp1, samp.val[i])
        val1 = (real_energy(fld.real) + imag_energy(fld.imag)).val
        val2 = complex_energy(fld).val
        np.testing.assert_allclose(val1, val2, atol=10./np.sqrt(Nsamp))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
from numpy.testing import assert_allclose

from ..common import list2fixture, setup_function, teardown_function

s = list2fixture([
    ift.RGSpace(8, distances=12.9),
    ift.RGSpace(59, distances=.24, harmonic=True),
    ift.RGSpace([12, 3])
])


def test_value(s):
    Regrid = ift.RegriddingOperator(s, s.shape)
    f = ift.from_random(Regrid.domain, 'normal')
    assert_allclose(f.val, Regrid(f).val)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..logger import logger
from .minimizer import Minimizer


class ConjugateGradient(Minimizer):
    """Implementation of the Conjugate Gradient scheme.

    It is an iterative method for solving a linear system of equations:
                                    Ax = b

    Parameters
    ----------
    controller : :py:class:`nifty8.IterationController`
        Object that decides when to terminate the minimization.
    nreset : int
        every `nreset` CG steps the residual will be recomputed accurately
        by applying the operator instead of updating the old residual

    References
    ----------
    Jorge Nocedal & Stephen Wright, "Numerical Optimization", Second Edition,
    2006, Springer-Verlag New York
    """

    def __init__(self, controller, nreset=20):
        self._controller = controller
        self._nreset = nreset

    def __call__(self, energy, preconditioner=None):
        """Runs the conjugate gradient minimization.

        Parameters
        ----------
        energy : Energy object at the starting point of the iteration.
            Its metric operator must be independent of position, otherwise
            linear conjugate gradient minimization will fail.
        preconditioner : Operator *optional*
            This operator can be provided which transforms the variables of the
            system to improve the conditioning. Default: None.

        Returns
        -------
        QuadraticEnergy
            state at last point of the iteration
        int
            Can be controller.CONVERGED or controller.ERROR
        """
        controller = self._controller
        status = controller.start(energy)
        if status != controller.CONTINUE:
            return energy, status

        r = energy.gradient
        d = r if preconditioner is None else preconditioner(r)

        previous_gamma = r.s_vdot(d).real
        if np.isnan(previous_gamma):
            logger.error("Error: ConjugateGradient: previous_gamma==NaN")
            return energy, controller.ERROR
        if previous_gamma == 0:
            return energy, controller.CONVERGED

        ii = 0
        while True:
            q = energy.apply_metric(d)
            curv = d.s_vdot(q).real
            if np.isnan(curv):
                logger.error("Error: ConjugateGradient: curv==NaN")
                return energy, controller.ERROR
            if curv == 0.:
                logger.error("Error: ConjugateGradient: curv==0.")
                return energy, controller.ERROR
            alpha = previous_gamma/curv

            if alpha < 0:
                logger.error("Error: ConjugateGradient: alpha<0.")
                return energy, controller.ERROR

            ii += 1
            if ii < self._nreset:
                r = r - q*alpha
                energy = energy.at_with_grad(energy.position - alpha*d, r)
            else:
                energy = energy.at(energy.position - alpha*d)
                r = energy.gradient
                ii = 0

            s = r if preconditioner is None else preconditioner(r)

            gamma = r.s_vdot(s).real
            if np.isnan(gamma):
                logger.error("Error: ConjugateGradient: gamma==NaN")
                return energy, controller.ERROR
            if gamma < 0:
                logger.error(
                    "Positive definiteness of preconditioner violated!")
                return energy, controller.ERROR
            if gamma == 0:
                return energy, controller.CONVERGED

            status = controller.check(energy)
            if status != controller.CONTINUE:
                return energy, status

            d = d * max(0, gamma/previous_gamma) + s

            previous_gamma = gamma



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society, Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from functools import reduce
from operator import add

import numpy as np

from .. import utilities
from ..domain_tuple import DomainTuple
from ..field import Field
from ..linearization import Linearization
from ..multi_domain import MultiDomain
from ..multi_field import MultiField
from ..sugar import makeDomain, makeOp
from ..utilities import iscomplextype, myassert
from .adder import Adder
from .linear_operator import LinearOperator
from .operator import Operator, _OpChain, _OpSum
from .sampling_enabler import SamplingEnabler
from .sandwich_operator import SandwichOperator
from .scaling_operator import ScalingOperator
from .simple_linear_operators import FieldAdapter, VdotOperator


def _check_sampling_dtype(domain, dtypes):
    if dtypes is None:
        return
    if isinstance(domain, DomainTuple):
        np.dtype(dtypes)
        return
    elif isinstance(domain, MultiDomain):
        if isinstance(dtypes, dict):
            for dt in dtypes.values():
                np.dtype(dt)
            if set(domain.keys()) == set(dtypes.keys()):
                return
        else:
            np.dtype(dtypes)
            return
    raise TypeError


def _field_to_dtype(field):
    if isinstance(field, Field):
        dt = field.dtype
    elif isinstance(field, MultiField):
        dt = {kk: ff.dtype for kk, ff in field.items()}
    else:
        raise TypeError
    _check_sampling_dtype(field.domain, dt)
    return dt


class EnergyOperator(Operator):
    """Operator which has a scalar domain as target domain.

    It is intended as an objective function for field inference.  It can
    implement a positive definite, symmetric form (called `metric`) that is
    used as curvature for second-order minimizations.

    Examples
    --------
     - Information Hamiltonian, i.e. negative-log-probabilities.
     - Gibbs free energy, i.e. an averaged Hamiltonian, aka Kullback-Leibler
       divergence.
    """
    _target = DomainTuple.scalar_domain()


class LikelihoodEnergyOperator(EnergyOperator):
    """Represent a negative log-likelihood.

    The input to the Operator are the parameters of the negative log-likelihood.
    Unlike a general `EnergyOperator`, the metric of a
    `LikelihoodEnergyOperator` is the Fisher information metric of the
    likelihood.
    """
    def __init__(self, data_residual, sqrt_data_metric_at):
        from ..extra import is_operator
        if data_residual is not None and not is_operator(data_residual):
            raise TypeError(f"{data_residual} is not an operator")
        self._res = data_residual
        self._sqrt_data_metric_at = sqrt_data_metric_at
        self._name = None

    def normalized_residual(self, x):
        return (self._sqrt_data_metric_at(x) @ self._res).force(x)

    @property
    def data_domain(self):
        if self._res is None:
            return None
        return self._res.target

    def get_transformation(self):
        """The coordinate transformation that maps into a coordinate system in
        which the metric of a likelihood is the Euclidean metric.

        Returns
        -------
        np.dtype, or dict of np.dtype : The dtype(s) of the target space of the
        transformation.

        Operator : The transformation that maps from `domain` into the
        Euclidean target space.

        Note
        ----
        This Euclidean target space is the disjoint union of the Euclidean
        target spaces of all summands. Therefore, the keys of `MultiDomains`
        are prefixed with an index and `DomainTuples` are converted to
        `MultiDomains` with the index as the key.
        """
        raise NotImplementedError

    def __matmul__(self, other):
        return _LikelihoodChain(self, other)

    def __rmatmul__(self, other):
        return _LikelihoodChain(other, self)

    def __add__(self, other):
        return _LikelihoodSum.make([self, other])

    def __radd__(self, other):
        return _LikelihoodSum.make([other, self])

    def get_metric_at(self, x):
        """Compute the Fisher information metric for a `LikelihoodEnergyOperator`
        at `x` using the Jacobian of the coordinate transformation given by
        :func:`~nifty8.operators.operator.Operator.get_transformation`. """
        dtp, f = self.get_transformation()
        bun = f(Linearization.make_var(x)).jac
        return SandwichOperator.make(bun, sampling_dtype=dtp)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, x):
        if isinstance(self, _LikelihoodSum):
            raise RuntimeError("The name of a LikelihoodSum cannot be set. "
                               "Set the name of each individual LikelihoodEnergy separately.")
        self._name = x


class _LikelihoodChain(LikelihoodEnergyOperator):
    def __init__(self, op1, op2):
        from .simple_linear_operators import PartialExtractor
        self._op = _OpChain.make((op1, op2))
        self._domain = self._op.domain

        if isinstance(op1, ScalingOperator):
            res = op2._res
            sqrt_data_metric_at = op2._sqrt_data_metric_at
        elif op1._res is None:
            res = None
            sqrt_data_metric_at = None
        else:
            if isinstance(op2.target, MultiDomain):
                extract = PartialExtractor(op2.target, op1._res.domain)
            else:
                extract = Operator.identity_operator(op2.target)
            res = op1._res @ extract @ op2
            sqrt_data_metric_at = lambda x: op1._sqrt_data_metric_at(op2.force(x))

        super(_LikelihoodChain, self).__init__(res, sqrt_data_metric_at)
        self.name = (op2 if isinstance(op1, ScalingOperator) else op1).name

    def get_transformation(self):
        scaled_lh = isinstance(self._op._ops[0], ScalingOperator)
        ii = 1 if scaled_lh else 0
        tr = self._op._ops[ii].get_transformation()
        if tr is None:
            return tr
        dtype, trafo = tr
        if scaled_lh:
            trafo = trafo.scale(np.sqrt(self._op._ops[0]._factor))
        return dtype, _OpChain.make((trafo,)+self._op._ops[ii+1:])

    def apply(self, x):
        self._check_input(x)
        return self._op(x)

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        return self._op._simplify_for_constant_input_nontrivial(c_inp)

    def __repr__(self):
        return self._op.__repr__()


class _LikelihoodSum(LikelihoodEnergyOperator):
    def __init__(self, ops, _callingfrommake=False):
        from .simple_linear_operators import PrependKey

        if not _callingfrommake:
            raise NotImplementedError
        if len(set([isinstance(oo.domain, DomainTuple) for oo in ops])) > 1:
            raise RuntimeError("Some operators have DomainTuple and others have "
                    "MultiDomain as domain. This should not happen.")
        self._ops = ops
        res, prep, data_ops = [], [], []
        for ii, oo in enumerate(ops):
            rr = oo._res
            if rr is None:
                continue
            tgt = oo.data_domain
            lprep = Operator.identity_operator(tgt)
            key = self._get_name(ii)
            if isinstance(lprep.target, DomainTuple):
                lprep = lprep.ducktape_left("")
            else:
                key = key + ": "
            lprep = PrependKey(lprep.target, key) @ lprep
            prep.append(lprep)
            res.append(lprep @ rr)
            data_ops.append(oo)

        def sqrt_data_metric_at(x):
            return reduce(add, (pp @ oo._sqrt_data_metric_at(x) @ pp.adjoint
                                for pp, oo in zip(prep, data_ops)))

        lst = self._all_names()
        if len(lst) != len(set(lst)):
            raise ValueError(f"Name collision in likelihoods detected: {lst}")

        data_residuals = reduce(add, res)
        super(_LikelihoodSum, self).__init__(data_residuals, sqrt_data_metric_at)
        self._domain = data_residuals.domain

    @classmethod
    def unpack(cls, ops, res):
        for op in ops:
            if isinstance(op, cls):
                res = cls.unpack(op._ops, res)
            else:
                res = res + [op]
        return res

    @classmethod
    def make(cls, ops):
        res = cls.unpack(ops, [])
        if len(res) == 1:
            return res[0]
        return cls(res, _callingfrommake=True)

    def apply(self, x):
        from ..linearization import Linearization
        self._check_input(x)
        return _OpSum._apply_operator_sum(x, self._ops)

    def get_transformation(self):
        from .simple_linear_operators import PrependKey

        if any(oo.get_transformation() is None for oo in self._ops):
            return None
        dtype, trafo = {}, []
        for ii, lh in enumerate(self._ops):
            dtp, tr = lh.get_transformation()
            key = self._get_name(ii)
            if isinstance(tr.target, MultiDomain):
                key = key + ": "
                dtype.update({key+d: dtp[d] for d in dtp.keys()})
                tr = PrependKey(tr.target, key) @ tr
            else:
                dtype[key] = dtp
                tr = tr.ducktape_left(key)
            trafo.append(tr)
        return dtype, reduce(add, trafo)

    def _all_names(self):
        return [self._get_name(ii) for ii in range(len(self._ops))]

    def _get_name(self, i):
        res = self._ops[i].name
        if res is None:
            return f"Likelihood {i}"
        return res

    def __repr__(self):
        subs = []
        for ii, oo in enumerate(self._ops):
            subs.append(f"*{self._get_name(ii)}*")
            subs.append(oo.__repr__())
        return "_LikelihoodSum:\n" + utilities.indent("\n".join(subs))


class Squared2NormOperator(EnergyOperator):
    """Computes the square of the L2-norm of the output of an operator.

    Parameters
    ----------
    domain : Domain, DomainTuple or tuple of Domain
        Domain of the operator in which the L2-norm shall be computed.
    """

    def __init__(self, domain):
        self._domain = domain

    def apply(self, x):
        self._check_input(x)
        if x.jac is None:
            return x.vdot(x)
        res = x.val.vdot(x.val)
        return x.new(res, VdotOperator(2*x.val))


class QuadraticFormOperator(EnergyOperator):
    """Computes the L2-norm of a Field or MultiField with respect to a
    specific kernel given by `endo`.

    .. math ::
        E(f) = \\frac12 f^\\dagger \\text{endo}(f)

    Parameters
    ----------
    endo : EndomorphicOperator
         Kernel of the quadratic form
    """

    def __init__(self, endo):
        from .endomorphic_operator import EndomorphicOperator
        if not isinstance(endo, EndomorphicOperator):
            raise TypeError("op must be an EndomorphicOperator")
        self._op = endo
        self._domain = endo.domain

    def apply(self, x):
        self._check_input(x)
        if x.jac is None:
            return 0.5*x.vdot(self._op(x))
        res = 0.5*x.val.vdot(self._op(x.val))
        return x.new(res, VdotOperator(self._op(x.val)))


class VariableCovarianceGaussianEnergy(LikelihoodEnergyOperator):
    """Computes the negative log pdf of a Gaussian with unknown covariance.

    The covariance is assumed to be diagonal.

    .. math ::
        E(s,D) = - \\log G(s, C) = 0.5 (s)^\\dagger C (s) - 0.5 tr log(C),

    an information energy for a Gaussian distribution with residual s and
    inverse diagonal covariance C.
    The domain of this energy will be a MultiDomain with two keys,
    the target will be the scalar domain.

    Parameters
    ----------
    domain : Domain, DomainTuple, tuple of Domain
        domain of the residual and domain of the covariance diagonal.

    residual_key : key
        Residual key of the Gaussian.

    inverse_covariance_key : key
        Inverse covariance diagonal key of the Gaussian.

    sampling_dtype : np.dtype
        Data type of the samples. Usually either 'np.float*' or 'np.complex*'

    use_full_fisher: boolean
        Determines if the proper Fisher information metric should be used as
        `metric`. If False, the same approximation as in `get_transformation`
        is used. Default is True.
    """

    def __init__(self, domain, residual_key, inverse_covariance_key,
                 sampling_dtype, use_full_fisher=True):
        self._kr = str(residual_key)
        self._ki = str(inverse_covariance_key)
        dom = DomainTuple.make(domain)
        self._domain = MultiDomain.make({self._kr: dom, self._ki: dom})
        self._dt = {self._kr: sampling_dtype, self._ki: np.float64}
        _check_sampling_dtype(self._domain, self._dt)
        self._cplx = iscomplextype(sampling_dtype)
        self._use_full_fisher = use_full_fisher
        super(VariableCovarianceGaussianEnergy, self).__init__(
                Operator.identity_operator(dom).ducktape(self._kr),
                lambda x: makeOp(x[self._ki].sqrt()))

    def apply(self, x):
        self._check_input(x)
        r, i = x[self._kr], x[self._ki]
        if self._cplx:
            res = 0.5*r.vdot(r*i.real).real - i.log().sum()
        else:
            res = 0.5*(r.vdot(r*i) - i.log().sum())
        if not x.want_metric:
            return res
        if not self._use_full_fisher:
            return res.add_metric(self.get_metric_at(x.val))
        fct = 1. if self._cplx else 0.5
        met = {self._kr: i.val, self._ki: fct*i.val**(-2)}
        met = MultiField.from_dict(met, domain=self._domain)
        met = makeOp(met, sampling_dtype=self._dt)
        return res.add_metric(met)

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        from .simplify_for_const import ConstantLikelihoodEnergyOperator
        myassert(len(c_inp.keys()) == 1)
        key = c_inp.keys()[0]
        myassert(key in self._domain.keys())
        cst = c_inp[key]
        if key == self._kr:
            res = _SpecialGammaEnergy(cst).ducktape(self._ki)
        else:
            icov = makeOp(cst, sampling_dtype=self._dt[self._kr])
            res = GaussianEnergy(data=None, inverse_covariance=icov).ducktape(self._kr)
            trlog = cst.log().sum().val_rw()
            if not self._cplx:
                trlog /= 2
            res = res + ConstantLikelihoodEnergyOperator(-trlog)
        res = res + ConstantLikelihoodEnergyOperator(0.)
        myassert(res.target is self.target)
        return None, res

    def get_transformation(self):
        """
        Note
        ----
        For `VariableCovarianceGaussianEnergy`, a global transformation to
        Euclidean space does not exist. A local approximation invoking the
        residual is used instead.
        """
        r = FieldAdapter(self._domain[self._kr], self._kr)
        ivar = FieldAdapter(self._domain[self._kr], self._ki).real
        sc = 1. if self._cplx else 0.5
        f = r.adjoint @ (ivar.sqrt()*r) + ivar.adjoint @ (sc*ivar.log())
        return self._dt, f


class _SpecialGammaEnergy(LikelihoodEnergyOperator):
    def __init__(self, residual):
        from .simplify_for_const import ConstantOperator

        self._domain = DomainTuple.make(residual.domain)
        self._resi = residual
        self._cplx = iscomplextype(self._resi.dtype)
        self._dt = self._resi.dtype
        super(_SpecialGammaEnergy, self).__init__(
                ConstantOperator(self._resi, domain=self._domain),
                lambda x: self.get_metric_at(x).get_sqrt())

    def apply(self, x):
        self._check_input(x)
        r = self._resi
        if self._cplx:
            res = 0.5*(r*x.real).vdot(r).real - x.log().sum()
        else:
            res = 0.5*((r*x).vdot(r) - x.log().sum())
        if not x.want_metric:
            return res
        return res.add_metric(self.get_metric_at(x.val))

    def get_transformation(self):
        sc = 1. if self._cplx else np.sqrt(0.5)
        return self._dt, Operator.identity_operator(self._domain).log().scale(sc)


class GaussianEnergy(LikelihoodEnergyOperator):
    """Computes a negative-log Gaussian.

    Represents up to constants in :math:`d`:

    .. math ::
        E(f) = - \\log G(f-d, D) = 0.5 (f-d)^\\dagger D^{-1} (f-d),

    an information energy for a Gaussian distribution with data d and
    covariance D.

    Parameters
    ----------
    data : :class:`nifty8.field.Field` or None
        Observed data of the Gaussian likelihood. If `inverse_covariance` is
        `None`, the `dtype` of `data` is used for sampling. Default is
        0.
    inverse_covariance : LinearOperator
        Inverse covariance of the Gaussian. Default is the identity operator.
    domain : Domain, DomainTuple, tuple of Domain or MultiDomain
        Operator domain. By default it is inferred from `data` or
        `covariance` if specified
    sampling_dtype : dtype or dict of dtype
        Type used for sampling from the inverse covariance if
        `inverse_covariance` and `data` is `None`. Otherwise, this parameter
        does not have an effect. Default: None.

    Note
    ----
    At least one of the arguments has to be provided.
    """

    def __init__(self, data=None, inverse_covariance=None, domain=None, sampling_dtype=None):
        from ..sugar import full

        if inverse_covariance is not None and not isinstance(inverse_covariance, LinearOperator):
            raise TypeError

        self._domain = self._parseDomain(data, inverse_covariance, domain)

        if not isinstance(data, (Field, MultiField)) and data is not None:
            raise TypeError

        self._icov = inverse_covariance
        if inverse_covariance is None:
            self._op = Squared2NormOperator(self._domain).scale(0.5)
            dt = sampling_dtype if data is None else data.dtype
            self._icov = ScalingOperator(self._domain, 1., dt)
        else:
            self._op = QuadraticFormOperator(inverse_covariance)
            self._icov = inverse_covariance

        self._data = data
        if data is None:
            res = Operator.identity_operator(self._domain)
        else:
            res = Adder(data, neg=True)
        super(GaussianEnergy, self).__init__(res, lambda x: self.get_metric_at(x).get_sqrt())

        icovdtype = self._icov.sampling_dtype
        if icovdtype is not None and data is not None and icovdtype != data.dtype:
            for i0, i1 in [(icovdtype, data.dtype), (data.dtype, icovdtype)]:
                if isinstance(i0, dict) and not isinstance(i1, dict):
                    fst = list(i0.values())[0]
                    if all(elem == fst for elem in i0.values()) and i1 == fst:
                        return
            s = "Sampling dtype of inverse covariance does not match dtype of data.\n"
            s += f"icov.sampling_dtype: {icovdtype}\n"
            s += f"data.dtype: {data.dtype}"
            raise RuntimeError(s)


    @staticmethod
    def _checkEquivalence(olddom, newdom):
        newdom = makeDomain(newdom)
        if olddom is None:
            return newdom
        utilities.check_object_identity(olddom, newdom)
        return newdom

    def _parseDomain(self, data, inverse_covariance, domain):
        dom = None
        if inverse_covariance is not None:
            dom = self._checkEquivalence(dom, inverse_covariance.domain)
        if data is not None:
            dom = self._checkEquivalence(dom, data.domain)
        if domain is not None:
            dom = self._checkEquivalence(dom, domain)
        if dom is None:
            raise ValueError("no domain given")
        return dom

    def apply(self, x):
        self._check_input(x)
        residual = x if self._data is None else x - self._data
        res = self._op(residual).real
        if x.want_metric:
            return res.add_metric(self.get_metric_at(x.val))
        return res

    def get_transformation(self):
        return self._icov.sampling_dtype, self._icov.get_sqrt()

    def __repr__(self):
        dom = '()' if isinstance(self.domain, DomainTuple) else self.domain.keys()
        return f'GaussianEnergy {dom}'


class PoissonianEnergy(LikelihoodEnergyOperator):
    """Computes likelihood Hamiltonians of expected count field constrained by
    Poissonian count data.

    Represents up to an f-independent term :math:`log(d!)`:

    .. math ::
        E(f) = -\\log \\text{Poisson}(d|f) = \\sum f - d^\\dagger \\log(f),

    where f is a :class:`Field` in data space with the expectation values for
    the counts.

    Parameters
    ----------
    d : :class:`nifty8.field.Field`
        Data field with counts. Needs to have integer dtype and all field
        values need to be non-negative.
    """

    def __init__(self, d):
        if not isinstance(d, Field) or not np.issubdtype(d.dtype, np.integer):
            te = "data is of invalid data-type; counts need to be integers"
            raise TypeError(te)
        if np.any(d.val < 0):
            ve = "count data is negative and thus can not be Poissonian"
            raise ValueError(ve)
        self._d = d
        self._domain = DomainTuple.make(d.domain)
        super(PoissonianEnergy, self).__init__(Adder(d, neg=True),
                                               lambda x: self.get_metric_at(x).get_sqrt())

    def apply(self, x):
        self._check_input(x)
        res = x.sum() - x.log().vdot(self._d)
        if not x.want_metric:
            return res
        return res.add_metric(self.get_metric_at(x.val))

    def get_transformation(self):
        return np.float64, 2.*Operator.identity_operator(self._domain).sqrt()


class InverseGammaEnergy(LikelihoodEnergyOperator):
    """Computes the negative log-likelihood of the inverse gamma distribution.

    It negative log-pdf(x) is given by

    .. math ::

        \\sum_i (\\alpha_i+1)*\\ln(x_i) + \\beta_i/x_i

    This is the likelihood for the variance :math:`x=S_k` given data
    :math:`\\beta = 0.5 |s_k|^2` where the :class:`nifty8.field.Field`
    :math:`s` is known to have the covariance :math:`S_k`.

    Parameters
    ----------
    beta : :class:`nifty8.field.Field`
        beta parameter of the inverse gamma distribution
    alpha : Scalar, :class:`nifty8.field.Field`, optional
        alpha parameter of the inverse gamma distribution
    """

    def __init__(self, beta, alpha=-0.5):
        from .simplify_for_const import ConstantOperator

        if not isinstance(beta, Field):
            raise TypeError
        self._domain = DomainTuple.make(beta.domain)
        self._beta = beta
        if np.isscalar(alpha):
            alpha = Field(beta.domain, np.full(beta.shape, alpha))
        elif not isinstance(alpha, Field):
            raise TypeError
        self._alphap1 = alpha+1
        if not self._beta.dtype == np.float64:
            # FIXME Add proper complex support for this energy
            raise TypeError
        self._sampling_dtype = _field_to_dtype(self._beta)

        super(InverseGammaEnergy, self).__init__(
                2*ConstantOperator(self._beta, domain=self._domain),
                lambda x: makeOp(x.reciprocal().sqrt()))

    def apply(self, x):
        self._check_input(x)
        res = x.log().vdot(self._alphap1) + x.reciprocal().vdot(self._beta)
        if not x.want_metric:
            return res
        return res.add_metric(self.get_metric_at(x.val))

    def get_transformation(self):
        fact = self._alphap1.sqrt()
        res = makeOp(fact) @ Operator.identity_operator(self._domain).log()
        return self._sampling_dtype, res


class StudentTEnergy(LikelihoodEnergyOperator):
    """Computes likelihood energy corresponding to Student's t-distribution.

    .. math ::
        E_\\theta(f) = -\\log \\text{StudentT}_\\theta(f)
                     = \\frac{\\theta + 1}{2} \\log(1 + \\frac{f^2}{\\theta}),

    where f is a field defined on `domain`. Assumes that the data is `float64`
    for sampling.

    Parameters
    ----------
    domain : `Domain` or `DomainTuple`
        Domain of the operator
    theta : Scalar or :class:`nifty8.field.Field`
        Degree of freedom parameter for the student t distribution
    """

    def __init__(self, domain, theta):
        self._domain = DomainTuple.make(domain)
        self._theta = theta
        inp = Operator.identity_operator(self._domain)
        super(StudentTEnergy, self).__init__(inp, lambda x: self.get_metric_at(x).get_sqrt())

    def apply(self, x):
        self._check_input(x)
        res = (((self._theta+1)/2)*(x**2/self._theta).log1p()).sum()
        if not x.want_metric:
            return res
        return res.add_metric(self.get_metric_at(x.val))

    def get_transformation(self):
        if isinstance(self._theta, Field) or isinstance(self._theta, MultiField):
            th = self._theta
        else:
            from ..sugar import full
            th = full(self._domain, self._theta)
        return np.float64, makeOp(((th+1)/(th+3)).sqrt())


class BernoulliEnergy(LikelihoodEnergyOperator):
    """Computes likelihood energy of expected event frequency constrained by
    event data.

    .. math ::
        E(f) = -\\log \\text{Bernoulli}(d|f)
             = -d^\\dagger \\log f  - (1-d)^\\dagger \\log(1-f),

    where f is a field defined on `d.domain` with the expected
    frequencies of events.

    Parameters
    ----------
    d : :class:`nifty8.field.Field`
        Data field with events (1) or non-events (0).
    """

    def __init__(self, d):
        if not isinstance(d, Field) or not np.issubdtype(d.dtype, np.integer):
            raise TypeError
        if np.any(np.logical_and(d.val != 0, d.val != 1)):
            raise ValueError
        self._d = d
        self._domain = DomainTuple.make(d.domain)
        super(BernoulliEnergy, self).__init__(Adder(d, neg=True),
                                              lambda x: self.get_metric_at(x).get_sqrt())

    def apply(self, x):
        self._check_input(x)
        res = -x.log().vdot(self._d) + (1.-x).log().vdot(self._d-1.)
        if not x.want_metric:
            return res
        return res.add_metric(self.get_metric_at(x.val))

    def get_transformation(self):
        from ..sugar import full
        res = Adder(full(self._domain, 1.)) @ ScalingOperator(self._domain, -1)
        res = res * Operator.identity_operator(self._domain).reciprocal()
        return np.float64, -2.*res.sqrt().arctan()


class StandardHamiltonian(EnergyOperator):
    """Computes an information Hamiltonian in its standard form, i.e. with the
    prior being a real-valued Gaussian with unit covariance.

    Let the likelihood energy be :math:`E_{lh}`. Then this operator computes:

    .. math ::
         H(f) = 0.5 f^\\dagger f + E_{lh}(f):

    Other field priors can be represented via transformations of a white
    Gaussian field into a field with the desired prior probability structure.

    By implementing prior information this way, the field prior is represented
    by a generative model, from which NIFTy can draw samples and infer a field
    using the Maximum a Posteriori (MAP) or the Variational Bayes (VB) method.

    The metric of this operator can be used as covariance for drawing Gaussian
    samples.

    Parameters
    ----------
    lh : EnergyOperator
        The likelihood energy.
    ic_samp : IterationController
        Tells an internal :class:`SamplingEnabler` which convergence criterion
        to use to draw Gaussian samples.

    See also
    --------
    `Encoding prior knowledge in the structure of the likelihood`,
    Jakob Knollmüller, Torsten A. Ensslin,
    `<https://arxiv.org/abs/1812.04403>`_
    """

    def __init__(self, lh, ic_samp=None):
        self._lh = lh
        self._prior = GaussianEnergy(data=None, domain=lh.domain, sampling_dtype=float)
        self._ic_samp = ic_samp
        self._domain = lh.domain

    def apply(self, x):
        self._check_input(x)
        lhx, prx = self._lh(x), self._prior(x)
        if not x.want_metric or self._ic_samp is None:
            return lhx + prx
        met = SamplingEnabler(lhx.metric, prx.metric, self._ic_samp)
        return (lhx+prx).add_metric(met)

    @property
    def prior_energy(self):
        return self._prior

    @property
    def likelihood_energy(self):
        return self._lh

    @property
    def iteration_controller(self):
        return self._ic_samp

    def __repr__(self):
        subs = 'Likelihood energy:\n{}'.format(utilities.indent(self._lh.__repr__()))
        subs += '\nPrior energy:\n{}'.format(self._prior)
        return 'StandardHamiltonian:\n' + utilities.indent(subs)

    def _simplify_for_constant_input_nontrivial(self, c_inp):
        out, lh1 = self._lh.simplify_for_constant_input(c_inp)
        return out, StandardHamiltonian(lh1, self._ic_samp)


class AveragedEnergy(EnergyOperator):
    """Averages an energy over samples.

    Parameters
    ----------
    h: Hamiltonian
       The energy to be averaged.
    res_samples : iterable of :class:`nifty8.field.Field`
       Set of residual sample points to be added to mean field for
       approximate estimation of the KL.

    Notes
    -----
    - Having symmetrized residual samples, with both :math:`v_i` and
      :math:`-v_i` being present, ensures that the distribution mean is
      exactly represented.

    - :class:`AveragedEnergy(h)` approximates
      :math:`\\left< H(f) \\right>_{G(f-m,D)}` if the residuals :math:`f-m`
      are drawn from a Gaussian distribution with covariance :math:`D`.
    """

    def __init__(self, h, res_samples):
        self._h = h
        self._domain = h.domain
        self._res_samples = tuple(res_samples)

    def apply(self, x):
        self._check_input(x)
        mymap = map(lambda v: self._h(x+v), self._res_samples)
        return utilities.my_sum(mymap)/len(self._res_samples)

    def get_transformation(self):
        dtp, trafo = self._h.get_transformation()
        mymap = map(lambda v: trafo@Adder(v), self._res_samples)
        return dtp, utilities.my_sum(mymap)/np.sqrt(len(self._res_samples))
